# -*- coding: utf-8 -*-
import matplotlib.ticker as ticker
import numpy as np


def get_target_diagram_axes(x, y, option):
    """
    Get axes value for target_diagram function.

    Determines the axes information for a target diagram given the axis values
    (X,Y) and the options in the data structure OPTION returned by the
    GET_TARGET_DIAGRAM_OPTIONS function.

    INPUTS:
    x      : values for x-axis
    y      : values for y-axis
    option : dictionary containing option values. (Refer to
             GET_TARGET_DIAGRAM_OPTIONS function for more information.)

    OUTPUTS:
    axes           : dictionary containing axes information for target diagram
    axes['xtick']  : x-values at which to place tick marks
    axes['ytick']  : y-values at which to place tick marks
    axes['xlabel'] : labels for xtick values
    axes['ylabel'] : labels for ytick values
    option : dictionary containing updated option values

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Nov 25, 2016

    @author: rochfordp
    """
    # Specify max/min for axes
    foundmax = 1 if option["axismax"] != 0.0 else 0
    if foundmax == 0:
        # Axis limit not specified
        maxx = np.amax(np.absolute(x))
        maxy = np.amax(np.absolute(y))
    else:
        # Axis limit is specified
        maxx = option["axismax"]
        maxy = option["axismax"]

    # Determine default number of tick marks
    xtickvals = ticker.AutoLocator().tick_values(-1.0 * maxx, maxx)
    ytickvals = ticker.AutoLocator().tick_values(-1.0 * maxy, maxy)
    ntest = np.sum(xtickvals > 0)
    if ntest > 0:
        nxticks = np.sum(xtickvals > 0)
        nyticks = np.sum(ytickvals > 0)

        # Save nxticks and nyticks as function attributes for later
        # retrieval in function calls
        get_target_diagram_axes.nxticks = nxticks
        get_target_diagram_axes.nyticks = nyticks
    else:
        # Use function attributes for nxticks and nyticks
        if hasattr(get_target_diagram_axes, "nxticks") and hasattr(
            get_target_diagram_axes, "nxticks"
        ):
            nxticks = get_target_diagram_axes.nxticks
            nyticks = get_target_diagram_axes.nyticks
        else:
            raise ValueError("No saved values for nxticks & nyticks.")

    # Set default tick increment and maximum axis values
    if foundmax == 0:
        maxx = xtickvals[-1]
        maxy = ytickvals[-1]
        option["axismax"] = max(maxx, maxy)

    # Check if equal axes requested
    if option["equalaxes"] == "on":
        if maxx > maxy:
            maxy = maxx
            nyticks = nxticks
        else:
            maxx = maxy
            nxticks = nyticks

    # Convert to integer if whole number
    if isinstance(maxx, float) and maxx.is_integer():
        maxx = int(round(maxx))
    if isinstance(maxx, float) and maxy.is_integer():
        maxy = int(round(maxy))
    minx = -maxx
    miny = -maxy

    # Determine tick values
    if len(option["ticks"]) > 0:
        xtick = option["ticks"]
        ytick = option["ticks"]
    else:
        tincx = maxx / nxticks
        tincy = maxy / nyticks
        xtick = np.arange(minx, maxx + tincx, tincx)
        ytick = np.arange(miny, maxy + tincy, tincy)

    # Assign tick label positions
    if len(option["xticklabelpos"]) == 0:
        option["xticklabelpos"] = xtick
    if len(option["yticklabelpos"]) == 0:
        option["yticklabelpos"] = ytick

    # Set tick labels using provided tick label positions
    xlabel = []
    ylabel = []

    # Set x tick labels
    for i, item in enumerate(xtick):
        index = np.where(option["xticklabelpos"] == item)
        if len(index) > 0:
            xlabel.append(str(item))
        else:
            xlabel.append("")

    # Set tick labels at 0 to blank
    index = np.where(abs(xtick) < 1.0e-7)
    index = np.asscalar(index[0])
    xlabel[index] = ""

    # Set y tick labels
    for i, item in enumerate(ytick):
        index = np.where(option["xticklabelpos"] == xtick[i])
        if len(index) > 0:
            ylabel.append(str(item))
        else:
            ylabel.append("")

    # Set tick labels at 0 to blank
    index = np.where(abs(ytick) < 1.0e-7)
    index = np.asscalar(index[0])
    ylabel[index] = ""

    # Store output variables in data structure
    axes = {}
    axes["xtick"] = xtick
    axes["ytick"] = ytick
    axes["xlabel"] = xlabel
    axes["ylabel"] = ylabel

    return axes



# -*- coding: utf-8 -*-

from unittest import TestCase

from tstoolbox import tsutils


class TestMakeList(TestCase):
    @staticmethod
    def test_make_list():
        assert tsutils.make_list(None) is None
        assert tsutils.make_list(1) == [1]
        assert tsutils.make_list(1.2) == [1.2]
        assert tsutils.make_list("") is None
        assert tsutils.make_list("1,2") == [1, 2]
        assert tsutils.make_list("2,") == [2, None]
        assert tsutils.make_list([1, 2, 3]) == [1, 2, 3]
        assert tsutils.make_list(["1", 2, "3"]) == [1, 2, 3]
        assert tsutils.make_list([1, "er", 3]) == [1, "er", 3]
        assert tsutils.make_list([1, "er", 3.3]) == [1, "er", 3.3]
        assert tsutils.make_list(["1", "er", "3"]) == [1, "er", 3]
        assert tsutils.make_list(["1", "", "5.6"]) == [1, None, 5.6]
        assert tsutils.make_list(["1", "None", "5.6"]) == [1, None, 5.6]
        assert tsutils.make_list("None") is None
        assert tsutils.make_list("") is None
        assert tsutils.make_list("1.1,2.2,") == [1.1, 2.2, None]



# -*- coding: utf-8 -*-
"""
Certify the developer has input all requirements for PR.

Situations tested:

* additions are reported in CHANGELOG.rst
"""
from pathlib import Path

folder = Path(__file__).resolve().parents[1]
changelog = Path(folder, "docs", "CHANGELOG.rst")

with open(changelog, "r") as fin:
    for line in fin:
        if line.startswith("v"):
            raise ValueError(
                "Please add a summary of your additions to docs/CHANGELOG.rst. "
                "As described in: https://python-project-skeleton.readthedocs.io"
                "/en/latest/contributing.html#update-changelog."
            )
        if line.startswith("*"):
            break



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy as np


def overlay_taylor_diagram_lines(axes, cax, option):
    """
    Overlay lines emanating from origin on a Taylor diagram.

    OVERLAY_TAYLOR_DIAGRAM_CIRCLES(AXES,CAX,OPTION)
    Plots lines emanating from origin to indicate correlation values (CORs)

    INPUTS:
    axes   : data structure containing axes information for target diagram
    cax    : handle for plot axes
    option : data structure containing option values. (Refer to
             GET_TAYLOR_DIAGRAM_OPTIONS function for more information.)
    option['colcor']   : CORs grid and tick labels color (Default: blue)
    option['showlabelscor'] : Show or not the CORRELATION tick labels
    option['stylecor'] : Linestyle of the CORs grid
    option['tickcor']  : CORs values to plot lines from origin
    option['widthcor'] : Line width of the CORs grid

    OUTPUTS:
    None.

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 3, 2016
    """
    # DRAW CORRELATION LINES EMANATING FROM THE ORIGIN:
    npanels = option["numberpanels"]
    corr = option["tickcor"][npanels - 1]
    th = np.arccos(corr)
    cst = np.cos(th)
    snt = np.sin(th)
    cs = np.append(-1.0 * cst, cst)
    sn = np.append(-1.0 * snt, snt)
    for i, val in enumerate(cs):
        plt.plot(
            [0, axes["rmax"] * cs[i]],
            [0, axes["rmax"] * sn[i]],
            linestyle=option["stylecor"],
            color=option["colcor"],
            linewidth=option["widthcor"],
        )

    # annotate the lines by correlation coefficient
    if option["showlabelscor"] == "on":
        rt = 1.05 * axes["rmax"]
        for i, cc in enumerate(corr):
            plt.text(
                rt * cst[i],
                rt * snt[i],
                str(round(cc, 2)),
                horizontalalignment="center",
                color=option["colcor"],
            )



# -*- coding: utf-8 -*-

import shlex
import subprocess
from unittest import TestCase

import pandas
import pytest
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils


class TestAggregate(TestCase):
    def setUp(self):
        dr = pandas.date_range("2011-01-01", periods=2, freq="D")

        ts = pandas.Series([2, 2], index=dr)
        self.aggregate_direct_mean = pandas.DataFrame(ts, columns=["Value::mean"])
        self.aggregate_direct_mean.index.name = "Datetime"
        self.aggregate_direct_mean = tsutils.memory_optimize(self.aggregate_direct_mean)

        ts = pandas.Series([48, 48], index=dr)
        self.aggregate_direct_sum = pandas.DataFrame(ts, columns=["Value::sum"])
        self.aggregate_direct_sum.index.name = "Datetime"
        self.aggregate_direct_sum = tsutils.memory_optimize(self.aggregate_direct_sum)

        self.aggregate_cli_mean = b"""Datetime,Value::mean
2011-01-01,2.0
2011-01-02,2.0
"""

        self.aggregate_cli_sum = b"""Datetime,Value::sum
2011-01-01,48
2011-01-02,48
"""

    def test_aggregate_direct_mean(self):
        """Test daily mean aggregation."""
        out = tstoolbox.aggregate(
            statistic="mean", groupby="daily", input_ts="tests/data_flat.csv"
        ).astype("Int64")
        assert_frame_equal(out, self.aggregate_direct_mean)

    def test_aggregate_direct_sum(self):
        """Test daily mean summation."""
        out = tstoolbox.aggregate(
            statistic="sum", groupby="daily", input_ts="tests/data_flat.csv"
        )
        assert_frame_equal(out, self.aggregate_direct_sum)

    def test_aggregate_cli_mean(self):
        """Test CLI mean, daily (by default) aggregation."""
        args = (
            "tstoolbox aggregate "
            '--statistic="mean" '
            '--input_ts="tests/data_flat.csv"'
        )
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()[0]
        self.assertEqual(out, self.aggregate_cli_mean)

    def test_aggregate_cli_sum(self):
        """Test CLI summation, daily (by default) aggregation."""
        args = (
            "tstoolbox aggregate "
            '--statistic="sum" '
            '--input_ts="tests/data_flat.csv"'
        )
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()[0]
        self.assertEqual(out, self.aggregate_cli_sum)


def test_aggregate_groupby():
    """Test API ValueError, groupby and agg_interval."""
    with pytest.raises(ValueError):
        _ = tstoolbox.aggregate(
            statistic="sum",
            groupby="D",
            agg_interval="D",
            input_ts="tests/data_flat.csv",
        )


def test_aggregate_ninterval_groupby():
    """Test API ValueError, ninterval and groupby."""
    with pytest.raises(ValueError):
        _ = tstoolbox.aggregate(
            statistic="sum", groupby="7D", ninterval=7, input_ts="tests/data_flat.csv"
        )


def test_aggregate_bad_statistic():
    """Test API statistic name."""
    with pytest.raises(ValueError):
        _ = tstoolbox.aggregate(
            statistic="camel", groupby="D", input_ts="tests/data_flat.csv"
        )


def test_aggregate_agg_interval():
    """Test API agg_interval."""
    with pytest.warns(UserWarning):
        _ = tstoolbox.aggregate(
            statistic="mean", agg_interval="D", input_ts="tests/data_flat.csv"
        )


def test_aggregate_ninterval():
    """Test API ninterval."""
    with pytest.warns(UserWarning):
        _ = tstoolbox.aggregate(
            statistic="mean", ninterval=7, input_ts="tests/data_flat.csv"
        )



# -*- coding: utf-8 -*-
"""A collection of functions used by tstoolbox, wdmtoolbox, ...etc."""

from __future__ import division, print_function

import bz2
import datetime
import gzip
import inspect
import os
import sys
from functools import reduce, wraps
from io import BytesIO, StringIO
from math import gcd
from string import Template
from textwrap import TextWrapper
from typing import Any, Callable, List, Optional, Tuple, Union
from urllib.parse import urlparse

import dateparser
import numpy as np
import numpy.typing as npt
import pandas as pd
import pint_pandas
import typic
from _io import TextIOWrapper
from numpy import int64, ndarray
from pandas._libs.tslibs.timestamps import Timestamp
from pandas.core.frame import DataFrame
from pandas.core.indexes.base import Index
from pandas.tseries.frequencies import to_offset
from scipy.stats.distributions import lognorm, norm
from tabulate import simple_separated_format
from tabulate import tabulate as tb

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


@typic.al
def error_wrapper(estr: str) -> str:
    """Wrap estr into error format used by toolboxes."""
    wrapper = TextWrapper(initial_indent="*   ", subsequent_indent="*   ")
    nestr = ["", "*"]
    for paragraph in estr.split("\n\n"):
        nestr.append("\n".join(wrapper.wrap(paragraph.strip())))
        nestr.append("*")
    nestr.append("")
    return "\n".join(nestr)


_CODES = {}
_CODES["SUB_D"]: {
    "N": "Nanoseconds",
    "U": "microseconds",
    "L": "miLliseconds",
    "S": "Secondly",
    "T": "minuTely",
    "H": "Hourly",
}
_CODES["DAILY"]: {
    "D": "calendar Day",
    "B": "Business day",
    "C": "Custom business day (experimental)",
}
_CODES["WEEKLY"]: {
    "W": "Weekly",
    "W-SUN": "Weekly frequency (SUNdays)",
    "W-MON": "Weekly frequency (MONdays)",
    "W-TUE": "Weekly frequency (TUEsdays)",
    "W-WED": "Weekly frequency (WEDnesdays)",
    "W-THU": "Weekly frequency (THUrsdays)",
    "W-FRI": "Weekly frequency (FRIdays)",
    "W-SAT": "Weekly frequency (SATurdays)",
}
_CODES["MONTH"]: {
    "M": "Month end",
    "MS": "Month Start",
    "BM": "Business Month end",
    "BMS": "Business Month Start",
    "CBM": "Custom Business Month end",
    "CBMS": "Custom Business Month Start",
}
_CODES["QUARTERLY"]: {
    "Q": "Quarter end",
    "Q-JAN": "Quarterly, quarter ends end of JANuary",
    "Q-FEB": "Quarterly, quarter ends end of FEBruary",
    "Q-MAR": "Quarterly, quarter ends end of MARch",
    "Q-APR": "Quarterly, quarter ends end of APRil",
    "Q-MAY": "Quarterly, quarter ends end of MAY",
    "Q-JUN": "Quarterly, quarter ends end of JUNe",
    "Q-JUL": "Quarterly, quarter ends end of JULy",
    "Q-AUG": "Quarterly, quarter ends end of AUGust",
    "Q-SEP": "Quarterly, quarter ends end of SEPtember",
    "Q-OCT": "Quarterly, quarter ends end of OCTober",
    "Q-NOV": "Quarterly, quarter ends end of NOVember",
    "Q-DEC": "Quarterly, quarter ends end of DECember",
    "QS": "Quarter Start",
    "QS-JAN": "Quarterly, quarter Starts end of JANuary",
    "QS-FEB": "Quarterly, quarter Starts end of FEBruary",
    "QS-MAR": "Quarterly, quarter Starts end of MARch",
    "QS-APR": "Quarterly, quarter Starts end of APRil",
    "QS-MAY": "Quarterly, quarter Starts end of MAY",
    "QS-JUN": "Quarterly, quarter Starts end of JUNe",
    "QS-JUL": "Quarterly, quarter Starts end of JULy",
    "QS-AUG": "Quarterly, quarter Starts end of AUGust",
    "QS-SEP": "Quarterly, quarter Starts end of SEPtember",
    "QS-OCT": "Quarterly, quarter Starts end of OCTober",
    "QS-NOV": "Quarterly, quarter Starts end of NOVember",
    "QS-DEC": "Quarterly, quarter Starts end of DECember",
    "BQ": "Business Quarter end",
    "BQS": "Business Quarter Start",
}
_CODES["ANNUAL"]: {
    "A": "Annual end",
    "A-JAN": "Annual, year ends end of JANuary",
    "A-FEB": "Annual, year ends end of FEBruary",
    "A-MAR": "Annual, year ends end of MARch",
    "A-APR": "Annual, year ends end of APRil",
    "A-MAY": "Annual, year ends end of MAY",
    "A-JUN": "Annual, year ends end of JUNe",
    "A-JUL": "Annual, year ends end of JULy",
    "A-AUG": "Annual, year ends end of AUGust",
    "A-SEP": "Annual, year ends end of SEPtember",
    "A-OCT": "Annual, year ends end of OCTober",
    "A-NOV": "Annual, year ends end of NOVember",
    "A-DEC": "Annual, year ends end of DECember",
    "AS": "Annual Start",
    "AS-JAN": "Annual, year Starts end of JANuary",
    "AS-FEB": "Annual, year Starts end of FEBruary",
    "AS-MAR": "Annual, year Starts end of MARch",
    "AS-APR": "Annual, year Starts end of APRil",
    "AS-MAY": "Annual, year Starts end of MAY",
    "AS-JUN": "Annual, year Starts end of JUNe",
    "AS-JUL": "Annual, year Starts end of JULy",
    "AS-AUG": "Annual, year Starts end of AUGust",
    "AS-SEP": "Annual, year Starts end of SEPtember",
    "AS-OCT": "Annual, year Starts end of OCTober",
    "AS-NOV": "Annual, year Starts end of NOVember",
    "AS-DEC": "Annual, year Starts end of DECember",
    "BA": "Business Annual end",
    "BA-JAN": "Business Annual, business year ends end of JANuary",
    "BA-FEB": "Business Annual, business year ends end of FEBruary",
    "BA-MAR": "Business Annual, business year ends end of MARch",
    "BA-APR": "Business Annual, business year ends end of APRil",
    "BA-MAY": "Business Annual, business year ends end of MAY",
    "BA-JUN": "Business Annual, business year ends end of JUNe",
    "BA-JUL": "Business Annual, business year ends end of JULy",
    "BA-AUG": "Business Annual, business year ends end of AUGust",
    "BA-SEP": "Business Annual, business year ends end of SEPtember",
    "BA-OCT": "Business Annual, business year ends end of OCTober",
    "BA-NOV": "Business Annual, business year ends end of NOVember",
    "BA-DEC": "Business Annual, business year ends end of DECember",
    "BAS": "Business Annual Start",
    "BS-JAN": "Business Annual Start, business year starts end of JANuary",
    "BS-FEB": "Business Annual Start, business year starts end of FEBruary",
    "BS-MAR": "Business Annual Start, business year starts end of MARch",
    "BS-APR": "Business Annual Start, business year starts end of APRil",
    "BS-MAY": "Business Annual Start, business year starts end of MAY",
    "BS-JUN": "Business Annual Start, business year starts end of JUNe",
    "BS-JUL": "Business Annual Start, business year starts end of JULy",
    "BS-AUG": "Business Annual Start, business year starts end of AUGust",
    "BS-SEP": "Business Annual Start, business year starts end of SEPtember",
    "BS-OCT": "Business Annual Start, business year starts end of OCTober",
    "BS-NOV": "Business Annual Start, business year starts end of NOVember",
    "BS-DEC": "Business Annual Start, business year starts end of DECember",
}

docstrings = {
    "por": r"""por
        The `por` keyword adjusts the operation of `start_date` and `end_date`
        If "False" (the default) choose the indices in the time-series between
        `start_date` and `end_date`.  If "True" and if `start_date` or
        `end_date` is outside of the time-series will extend the time-series by
        inserting the exterior date(s).""",
    "lat": r"""lat
        The latitude of the point. North hemisphere is positive from 0 to 90. South
        hemisphere is negative from 0 to -90.""",
    "lon": r"""lon
        The longitude of the point.  Western hemisphere (west of Greenwich Prime
        Meridian) is negative 0 to -180.  The eastern hemisphere (east of the Greenwich
        Prime Meridian) is positive 0 to 180.""",
    "target_units": r"""target_units: str
        [optional, default is None, transformation]

        The purpose of this option is to specify target units for unit
        conversion.  The source units are specified in the header line
        of the input or using the 'source_units' keyword.

        The units of the input time-series or values are specified as
        the second field of a ':' delimited name in the header line of
        the input or in the 'source_units' keyword.

        Any unit string compatible with the 'pint' library can be used.

        This option will also add the 'target_units' string to the
        column names.""",
    "source_units": r"""source_units: str
        [optional, default is None, transformation]

        If unit is specified for the column as the second field of a ':'
        delimited column name, then the specified units and the
        'source_units' must match exactly.

        Any unit string compatible with the 'pint' library can be
        used.""",
    "names": r"""names: str
        [optional, default is None, transformation]

        If None, the column names are taken from the first row after
        'skiprows' from the input dataset.

        MUST include a name for all columns in the input dataset,
        including the index column.""",
    "index_type": r"""index_type : str
        [optional, default is 'datetime', output format]

        Can be either 'number' or 'datetime'.  Use 'number' with index
        values that are Julian dates, or other epoch reference.""",
    "input_ts": r"""input_ts : str
        [optional though required if using within Python, default is '-'
        (stdin)]

        Whether from a file or standard input, data requires a header of
        column names.  The default header is the first line of the
        input, but this can be changed using the 'skiprows' option.

        Most separators will be automatically detected. Most common date
        formats can be used, but the closer to ISO 8601 date/time
        standard the better.

        Command line:

            +-------------------------+------------------------+
            | Keyword Example         | Description            |
            +=========================+========================+
            | --input_ts=filename.csv | to read 'filename.csv' |
            +-------------------------+------------------------+
            | --input_ts='-'          | to read from standard  |
            |                         | input (stdin)          |
            +-------------------------+------------------------+

            In many cases it is better to use redirection rather that
            use `--input_ts=filename.csv`.  The following are identical:

            From a file:

                command subcmd --input_ts=filename.csv

            From standard input:

                command subcmd --input_ts=- < filename.csv

            The BEST way since you don't have to include `--input_ts=-`
            because that is the default:

                command subcmd < file.csv

            Can also combine commands by piping:

                command subcmd < filein.csv | command subcmd1 > fileout.csv

        As Python Library::

            You MUST use the `input_ts=...` option where `input_ts` can
            be one of a [pandas DataFrame, pandas Series, dict, tuple,
            list, StringIO, or file name].

            If result is a time series, returns a pandas DataFrame.""",
    "columns": r"""columns
        [optional, defaults to all columns, input filter]

        Columns to select out of input.  Can use column names from the
        first line header or column numbers.  If using numbers, column
        number 1 is the first data column.  To pick multiple columns;
        separate by commas with no spaces. As used in `tstoolbox pick`
        command.

        This solves a big problem so that you don't have to create
        a data set with a certain column order, you can rearrange
        columns when data is read in.""",
    "start_date": r"""start_date : str
        [optional, defaults to first date in time-series, input filter]

        The start_date of the series in ISOdatetime format, or 'None'
        for beginning.""",
    "end_date": r"""end_date : str
        [optional, defaults to last date in time-series, input filter]

        The end_date of the series in ISOdatetime format, or 'None' for
        end.""",
    "dropna": r"""dropna : str
        [optional, defauls it 'no', input filter]

        Set `dropna` to 'any' to have records dropped that have NA value
        in any column, or 'all' to have records dropped that have NA in
        all columns.  Set to 'no' to not drop any records.  The default
        is 'no'.""",
    "print_input": r"""print_input
        [optional, default is False, output format]

        If set to 'True' will include the input columns in the output
        table.""",
    "round_index": r"""round_index
        [optional, default is None which will do nothing to the index,
        output format]

        Round the index to the nearest time point.  Can significantly
        improve the performance since can cut down on memory and
        processing requirements, however be cautious about rounding to
        a very course interval from a small one.  This could lead to
        duplicate values in the index.""",
    "float_format": r"""float_format
        [optional, output format]

        Format for float numbers.""",
    "tablefmt": r"""tablefmt : str
        [optional, default is 'csv', output format]

        The table format.  Can be one of 'csv', 'tsv', 'plain',
        'simple', 'grid', 'pipe', 'orgtbl', 'rst', 'mediawiki', 'latex',
        'latex_raw' and 'latex_booktabs'.""",
    "header": r"""header : str
        [optional, default is 'default', output format]

        This is if you want a different header than is the default for
        this output table.  Pass a list of strings for each column in
        the table.""",
    "pandas_offset_codes": r"""+-------+---------------+
        | Alias | Description   |
        +=======+===============+
        | N     | Nanoseconds   |
        +-------+---------------+
        | U     | microseconds  |
        +-------+---------------+
        | L     | milliseconds  |
        +-------+---------------+
        | S     | Secondly      |
        +-------+---------------+
        | T     | Minutely      |
        +-------+---------------+
        | H     | Hourly        |
        +-------+---------------+
        | D     | calendar Day  |
        +-------+---------------+
        | W     | Weekly        |
        +-------+---------------+
        | M     | Month end     |
        +-------+---------------+
        | MS    | Month Start   |
        +-------+---------------+
        | Q     | Quarter end   |
        +-------+---------------+
        | QS    | Quarter Start |
        +-------+---------------+
        | A     | Annual end    |
        +-------+---------------+
        | AS    | Annual Start  |
        +-------+---------------+

        Business offset codes.

        +-------+------------------------------------+
        | Alias | Description                        |
        +=======+====================================+
        | B     | Business day                       |
        +-------+------------------------------------+
        | BM    | Business Month end                 |
        +-------+------------------------------------+
        | BMS   | Business Month Start               |
        +-------+------------------------------------+
        | BQ    | Business Quarter end               |
        +-------+------------------------------------+
        | BQS   | Business Quarter Start             |
        +-------+------------------------------------+
        | BA    | Business Annual end                |
        +-------+------------------------------------+
        | BAS   | Business Annual Start              |
        +-------+------------------------------------+
        | C     | Custom business day (experimental) |
        +-------+------------------------------------+
        | CBM   | Custom Business Month end          |
        +-------+------------------------------------+
        | CBMS  | Custom Business Month Start        |
        +-------+------------------------------------+

        Weekly has the following anchored frequencies:

        +-------+-------------+-------------------------------+
        | Alias | Equivalents | Description                   |
        +=======+=============+===============================+
        | W-SUN | W           | Weekly frequency (SUNdays)    |
        +-------+-------------+-------------------------------+
        | W-MON |             | Weekly frequency (MONdays)    |
        +-------+-------------+-------------------------------+
        | W-TUE |             | Weekly frequency (TUEsdays)   |
        +-------+-------------+-------------------------------+
        | W-WED |             | Weekly frequency (WEDnesdays) |
        +-------+-------------+-------------------------------+
        | W-THU |             | Weekly frequency (THUrsdays)  |
        +-------+-------------+-------------------------------+
        | W-FRI |             | Weekly frequency (FRIdays)    |
        +-------+-------------+-------------------------------+
        | W-SAT |             | Weekly frequency (SATurdays)  |
        +-------+-------------+-------------------------------+

        Quarterly frequencies (Q, BQ, QS, BQS) and annual frequencies
        (A, BA, AS, BAS) replace the "x" in the "Alias" column to have
        the following anchoring suffixes:

        +-------+----------+-------------+----------------------------+
        | Alias | Examples | Equivalents | Description                |
        +=======+==========+=============+============================+
        | x-DEC | A-DEC    | A           | year ends end of DECember  |
        |       | Q-DEC    | Q           |                            |
        |       | AS-DEC   | AS          |                            |
        |       | QS-DEC   | QS          |                            |
        +-------+----------+-------------+----------------------------+
        | x-JAN |          |             | year ends end of JANuary   |
        +-------+----------+-------------+----------------------------+
        | x-FEB |          |             | year ends end of FEBruary  |
        +-------+----------+-------------+----------------------------+
        | x-MAR |          |             | year ends end of MARch     |
        +-------+----------+-------------+----------------------------+
        | x-APR |          |             | year ends end of APRil     |
        +-------+----------+-------------+----------------------------+
        | x-MAY |          |             | year ends end of MAY       |
        +-------+----------+-------------+----------------------------+
        | x-JUN |          |             | year ends end of JUNe      |
        +-------+----------+-------------+----------------------------+
        | x-JUL |          |             | year ends end of JULy      |
        +-------+----------+-------------+----------------------------+
        | x-AUG |          |             | year ends end of AUGust    |
        +-------+----------+-------------+----------------------------+
        | x-SEP |          |             | year ends end of SEPtember |
        +-------+----------+-------------+----------------------------+
        | x-OCT |          |             | year ends end of OCTober   |
        +-------+----------+-------------+----------------------------+
        | x-NOV |          |             | year ends end of NOVember  |
        +-------+----------+-------------+----------------------------+""",
    "plotting_position_table": r"""+------------+--------+----------------------+-----------------------+
        | Name       | a      | Equation             | Description           |
        |            |        | (i-a)/(n+1-2*a)      |                       |
        +============+========+======================+=======================+
        | weibull    | 0      | i/(n+1)              | mean of sampling      |
        | (default)  |        |                      | distribution          |
        +------------+--------+----------------------+-----------------------+
        | benard     | 0.3    | (i-0.3)/(n+0.4)      | approx. median of     |
        |            |        |                      | sampling distribution |
        +------------+--------+----------------------+-----------------------+
        | filliben   | 0.3175 | (i-0.3175)/(n+0.365) |                       |
        +------------+--------+----------------------+-----------------------+
        | yu         | 0.326  | (i-0.326)/(n+0.348)  |                       |
        +------------+--------+----------------------+-----------------------+
        | tukey      | 1/3    | (i-1/3)/(n+1/3)      | approx. median of     |
        |            |        |                      | sampling distribution |
        +------------+--------+----------------------+-----------------------+
        | blom       | 0.375  | (i-0.375)/(n+0.25)   |                       |
        +------------+--------+----------------------+-----------------------+
        | cunnane    | 2/5    | (i-2/5)/(n+1/5)      | subjective            |
        +------------+--------+----------------------+-----------------------+
        | gringorton | 0.44   | (1-0.44)/(n+0.12)    |                       |
        +------------+--------+----------------------+-----------------------+
        | hazen      | 1/2    | (i-1/2)/n            | midpoints of n equal  |
        |            |        |                      | intervals             |
        +------------+--------+----------------------+-----------------------+
        | larsen     | 0.567  | (i-0.567)/(n-0.134)  |                       |
        +------------+--------+----------------------+-----------------------+
        | gumbel     | 1      | (i-1)/(n-1)          | mode of sampling      |
        |            |        |                      | distribution          |
        +------------+--------+----------------------+-----------------------+
        | california | NA     | i/n                  |                       |
        +------------+--------+----------------------+-----------------------+

        Where 'i' is the sorted rank of the y value, and 'n' is the
        total number of values to be plotted.""",
    "clean": r"""clean
        [optional, default is False, input filter]

        The 'clean' command will repair a input index, removing
        duplicate index values and sorting.""",
    "skiprows": r"""skiprows: list-like or integer or callable
        [optional, default is None which will infer header from first
        line, input filter]

        Line numbers to skip (0-indexed) or number of lines to skip
        (int) at the start of the file.

        If callable, the callable function will be evaluated against the
        row indices, returning True if the row should be skipped and
        False otherwise.  An example of a valid callable argument would
        be

        ``lambda x: x in [0, 2]``.""",
    "groupby": r"""groupby: str
        [optional, default is None, transformation]

        The pandas offset code to group the time-series data into.
        A special code is also available to group 'months_across_years'
        that will group into twelve monthly categories across the entire
        time-series.""",
    "force_freq": r"""force_freq: str
        [optional, output format]

        Force this frequency for the output.  Typically you will only
        want to enforce a smaller interval where tstoolbox will insert
        missing values as needed.  WARNING: you may lose data if not
        careful with this option.  In general, letting the algorithm
        determine the frequency should always work, but this option will
        override.  Use PANDAS offset codes.""",
    "output_names": r"""output_names: str
        [optional, output_format]

        The tstoolbox will change the names of the output columns to include
        some record of the operations used on each column.  The `output_names`
        will override that feature.  Must be a list or tuple equal to the
        number of columns in the output data.""",
}

# Decided this was inelegant, but left here in case I figure out what I want
# and how I want it.
# ntables = {}
# for key in ["SUB_D", "DAILY", "WEEKLY", "QUATERLY", "ANNUAL"]:
#     ntables[key] = tb(_CODES[key].items(),
#                       tablefmt="grid",
#                       headers=["Alias", "Description"],)
#     ntables[key] = "        ".join(ntables[key].splitlines(True))
# codes_table = f"""{ntables["SUB_D"]}
#
#     {ntables["DAILY"]}
#     """
#
# docstrings["pandas_offset_codes"] = codes_table


@typic.constrained(gt=0, lt=1)
class FloatBetweenZeroAndOne(float):
    """0.0 < float < 1.0"""


@typic.constrained(ge=0, le=1)
class FloatBetweenZeroAndOneInclusive(float):
    """0.0 <= float <= 1.0"""


@typic.constrained(ge=0)
class FloatGreaterEqualToZero(float):
    """float >= 0.0"""


@typic.constrained(ge=1)
class FloatGreaterEqualToOne(float):
    """float >= 1.0"""


@typic.constrained(ge=0)
class IntGreaterEqualToZero(int):
    """int >= 0"""


@typic.constrained(ge=1)
class IntGreaterEqualToOne(int):
    """int >= 1"""


@typic.constrained(ge=1, le=3)
class IntBetweenOneAndThree(int):
    """1 <= int <= 3"""


def flatten(list_of_lists):
    if isinstance(list_of_lists, (list, tuple)):
        if len(list_of_lists) == 0:
            return list_of_lists
        if isinstance(list_of_lists[0], (list, tuple)):
            return list(flatten(list_of_lists[0])) + list(flatten(list_of_lists[1:]))
        return list(list_of_lists[:1]) + list(flatten(list_of_lists[1:]))
    return list_of_lists


@typic.al
def stride_and_unit(sunit: str) -> Tuple[str, int]:
    """Split a stride/unit combination into component parts."""
    if sunit is None:
        return sunit
    unit = sunit.lstrip("+-. 1234567890")
    stride = sunit[: sunit.index(unit)]
    if len(stride) > 0:
        stride = int(stride)
    else:
        stride = 1
    return unit, stride


@typic.al
def set_ppf(ptype: Optional[Literal["norm", "lognorm", "weibull"]]) -> Callable:
    """Return correct Percentage Point Function for `ptype`."""
    if ptype == "norm":
        ppf = norm.ppf
    elif ptype == "lognorm":
        ppf = lognorm.freeze(0.5, loc=0).ppf
    elif ptype == "weibull":

        def ppf(y: Union[List[float], npt.ArrayLike]) -> npt.ArrayLike:
            """Percentage Point Function for the weibull distibution."""
            return np.log(-np.log(1 - np.array(y)))

    elif ptype is None:

        def ppf(y: npt.ArrayLike) -> npt.ArrayLike:
            return y

    return ppf


PPDICT = {
    "weibull": 0,
    "benard": 0.3,
    "filliben": 0.3175,
    "yu": 0.326,
    "tukey": 1 / 3,
    "blom": 0.375,
    "cunnane": 2 / 5,
    "gringorton": 0.44,
    "hazen": 1 / 2,
    "larsen": 0.567,
    "gumbel": 1,
}


@typic.al
def set_plotting_position(
    n: Union[int, int64],
    plotting_position: Union[
        float,
        Literal[
            "weibull",
            "benard",
            "bos-levenbach",
            "filliben",
            "yu",
            "tukey",
            "blom",
            "cunnane",
            "gringorton",
            "hazen",
            "larsen",
            "gumbel",
            "california",
        ],
    ] = "weibull",
) -> ndarray:
    """Create plotting position 1D array using linspace."""
    if plotting_position == "california":
        return np.linspace(1.0 / n, 1.0, n)
    a = PPDICT.get(plotting_position, plotting_position)
    i = np.arange(1, n + 1)
    return (i - a) / float(n + 1 - 2 * a)


@typic.al
def _handle_curly_braces_in_docstring(s: str, **kwargs) -> str:
    """Replace missing keys with a pattern."""
    RET = "{{{}}}"
    try:
        return s.format(**kwargs)
    except KeyError as e:
        keyname = e.args[0]
        return _handle_curly_braces_in_docstring(
            s, **{keyname: RET.format(keyname)}, **kwargs
        )


@typic.al
def copy_doc(source: Callable) -> Callable:
    """Copy docstring from source.

    Parameters
    ----------
    source : Callable
        Function to take doc string from.

    Examples
    --------
    >>> @decorit.copy_doc(func_to_copy_docstring_from)
    >>> def func(args):
    ...     pass  # function goes here
    # Function uses parameters of given func

    """

    @wraps(source)
    def wrapper_copy_doc(func: Callable) -> Callable:
        if source.__doc__:
            func.__doc__ = source.__doc__  # noqa: WPS125
        return func

    return wrapper_copy_doc


@typic.al
def doc(fdict: dict, **kwargs) -> Callable:
    """Return a decorator that formats a docstring."""

    def f(fn):
        fn.__doc__ = Template(fn.__doc__).safe_substitute(**fdict)

        # kwargs is currently always empty.
        # Could remove, but keeping in case useful in future.
        for attr in kwargs:
            setattr(fn, attr, kwargs[attr])
        return fn

    return f


@typic.al
def parsedate(
    dstr: Optional[str], strftime: Optional[Any] = None, settings: Optional[Any] = None
) -> Timestamp:
    """Use dateparser to parse a wide variety of dates.

    Used for start and end dates.
    """
    if dstr is None:
        return dstr

    # The API should boomerang a datetime.datetime instance and None.
    if isinstance(dstr, datetime.datetime):
        if strftime is None:
            return dstr
        return dstr.strftime(strftime)

    try:
        pdate = pd.to_datetime(dstr)
    except ValueError:
        pdate = dateparser.parse(dstr, settings=settings)

    if pdate is None:
        raise ValueError(
            error_wrapper(
                """
Could not parse date string '{}'.
""".format(
                    dstr
                )
            )
        )

    if strftime is None:
        return pdate
    return pdate.strftime(strftime)


@typic.al
def merge_dicts(*dict_args: dict) -> dict:
    """Merge multiple dictionaries."""
    result = {}
    for d in dict_args:
        result.update(d)
    return result


def about(name):
    """Return generic 'about' information used across all toolboxes."""
    import platform

    import pkg_resources

    namever = str(pkg_resources.get_distribution(name.split(".")[0]))
    print("package name = {}\npackage version = {}".format(*namever.split()))

    print("platform architecture = {}".format(platform.architecture()))
    print("platform machine = {}".format(platform.machine()))
    print("platform = {}".format(platform.platform()))
    print("platform processor = {}".format(platform.processor()))
    print("platform python_build = {}".format(platform.python_build()))
    print("platform python_compiler = {}".format(platform.python_compiler()))
    print("platform python branch = {}".format(platform.python_branch()))
    print(
        "platform python implementation = {}".format(platform.python_implementation())
    )
    print("platform python revision = {}".format(platform.python_revision()))
    print("platform python version = {}".format(platform.python_version()))
    print("platform release = {}".format(platform.release()))
    print("platform system = {}".format(platform.system()))
    print("platform version = {}".format(platform.version()))


def _round_index(ntsd: DataFrame, round_index: Optional[str] = None) -> DataFrame:
    """Round the index, typically time, to the nearest interval."""
    if round_index is None:
        return ntsd
    ntsd.index = ntsd.index.round(round_index)
    return ntsd


def _pick_column_or_value(tsd, var):
    """Return a keyword value or a time-series."""
    try:
        var = np.array([float(var)])
    except ValueError:
        var = tsd.loc[:, var].values
    return var


def make_list(*strorlist, **kwds: Any) -> Any:
    """Normalize strings, converting to numbers or lists."""
    try:
        n = kwds.pop("n")
    except KeyError:
        n = None
    if n is not None:
        n = int(n)

    try:
        sep = kwds.pop("sep")
    except KeyError:
        sep = ","

    try:
        kwdname = kwds.pop("kwdname")
    except KeyError:
        kwdname = ""

    try:
        flat = kwds.pop("flat")
    except KeyError:
        flat = True

    if isinstance(strorlist, (list, tuple)):
        # The following will fix ((tuples, in, a, tuple, problem),)
        if flat is True:
            strorlist = flatten(strorlist)

        if len(strorlist) == 1:
            # Normalize lists and tuples of length 1 to scalar for
            # further processing.
            strorlist = strorlist[0]

    if isinstance(strorlist, (list, tuple)) and n is not None and len(strorlist) != n:
        raise ValueError(
            error_wrapper(
                """
The list {0} for "{2}" should have {1} members according to function requirements.
""".format(
                    strorlist, n, kwdname
                )
            )
        )

    if isinstance(strorlist, pd.DataFrame):
        return [strorlist]

    if isinstance(strorlist, pd.Series):
        return [pd.DataFrame(strorlist)]

    if strorlist is None or isinstance(strorlist, (type(None))):
        # None -> None
        #
        return None

    if isinstance(strorlist, (int, float)):
        # 1      -> [1]
        # 1.2    -> [1.2]
        #
        return [strorlist]

    if isinstance(strorlist, (str, bytes)) and (
        strorlist in ["None", "", b"None", b""]
    ):
        # 'None' -> None
        # ''     -> None
        #
        return None

    if isinstance(strorlist, (str, bytes)):
        # '1'   -> [1]
        # '5.7' -> [5.7]

        # Anything other than a scalar int or float continues.
        #
        try:
            return [int(strorlist)]
        except ValueError:
            try:
                return [float(strorlist)]
            except ValueError:
                pass
        # Deal with a str or bytes.
        strorlist = strorlist.strip()

        if isinstance(strorlist, str):
            if "\r" in strorlist or "\n" in strorlist:
                return [StringIO(strorlist)]
            strorlist = strorlist.split(str(sep))

        if isinstance(strorlist, bytes):
            if b"\r" in strorlist or b"\n" in strorlist:
                return [BytesIO(strorlist)]
            strorlist = strorlist.split(bytes(sep, encoding="utf8"))

    if isinstance(strorlist, (StringIO, BytesIO)):
        return strorlist

    if n is None:
        n = len(strorlist)

    # At this point 'strorlist' variable should be a list or tuple.
    if len(strorlist) != n:
        raise ValueError(
            error_wrapper(
                """
The list {0} for "{2}" should have {1} members according to function requirements.
""".format(
                    strorlist, n, kwdname
                )
            )
        )

    # [1, 2, 3]          -> [1, 2, 3]
    # ['1', '2']         -> [1, 2]

    # [1, 'er', 5.6]     -> [1, 'er', 5.6]
    # [1,'er',5.6]       -> [1, 'er', 5.6]
    # ['1','er','5.6']   -> [1, 'er', 5.6]

    # ['1','','5.6']     -> [1, None, 5.6]
    # ['1','None','5.6'] -> [1, None, 5.6]

    ret = []
    for each in strorlist:
        if isinstance(each, (type(None), int, float, pd.DataFrame, pd.Series)):
            ret.append(each)
            continue
        if flat is False and isinstance(each, list):
            ret.append(each)
            continue
        if each is None or each.strip() == "" or each == "None":
            ret.append(None)
            continue
        try:
            ret.append(int(each))
        except ValueError:
            try:
                ret.append(float(each))
            except ValueError:
                ret.append(each)
    return ret


def make_iloc(columns, col_list):
    """Imitates the .ix option with subtracting one to convert."""
    # ["1", "Value2"]    ->    [0, "Value2"]
    # [1, 2, 3]          ->    [0, 1, 2]
    col_list = make_list(col_list)
    ntestc = []
    for i in col_list:
        try:
            ntestc.append(int(i) - 1)
        except ValueError:
            ntestc.append(columns.index(i))
    return ntestc


# NOT SET YET...
#
# Take `air_pressure` from df.loc[:, 1]
# Take `short_wave_rad` from df.loc[:, 'swaverad']
# The `temperature` keyword is set to 23.4 for all time periods
# The `wind_speed` keyword is set to 2.4 and 3.1 in turn
#
# Will output two columns, one with wind_speed equal to 2.4, the next
# with wind_speed equal to 3.1.
#
# API:
# testfunction(input_ts=df,
#              air_pressure='_1',
#              short_wave_rad='swaverad',
#              temperature=23.4,
#              wind_speed=[2.4, 3.1])
#             )
#
# CLI:
# mettoolbox testfunction --air_pressure=_1 \
#                         --short_wave_rad=swaverad \
#                         --temperature 23.4 \
#                         --wind_speed 2.4,3.1 < df.csv


def _normalize_units(
    ntsd: DataFrame,
    source_units: Optional[str],
    target_units: Optional[str],
    source_units_required: bool = False,
) -> DataFrame:
    """
    Following is aspirational and may not reflect the code.

    +--------------+--------------+--------------+--------------+--------------+
    | INPUT        | INPUT        | INPUT        | RETURN       | RETURN       |
    | ntsd.columns | source_units | target_units | source_units | target_units |
    +==============+==============+==============+==============+==============+
    | ["col1:cm",  | ["ft",       | ["m",        | ValueError   |              |
    |  "col2:cm"]  |  "cm"]       |  "cm"]       |              |              |
    +--------------+--------------+--------------+--------------+--------------+
    | ["col1:cm",  | ["cm"]       | ["ft"]       | ValueError   |              |
    |  "col2:cm"]  |              |              |              |              |
    +--------------+--------------+--------------+--------------+--------------+
    | ["col1:cm",  | ["cm"]       | ["ft"]       | ["cm",       | ["ft",       |
    |  "col2"]     |              |              |  ""]         |  ""]         |
    +--------------+--------------+--------------+--------------+--------------+
    | ["col1",     | ["", "cm"]   | ["", "ft"]   | ["",          | ["",         |
    |  "col2:cm"]  |              |              |  "cm"]       |  "ft"]       |
    +--------------+--------------+--------------+--------------+--------------+
    |              | ["cm"]       | ["ft"]       | ["cm"]       | ["ft"]       |
    +--------------+--------------+--------------+--------------+--------------+
    | ["cm"]       | None         | ["ft"]       | ["cm"]       | ["ft"]       |
    +--------------+--------------+--------------+--------------+--------------+

    """
    # Enforce DataFrame
    ntsd = pd.DataFrame(ntsd)

    target_units = make_list(target_units, n=len(ntsd.columns))
    if target_units is not None:
        target_units = ["" if i is None else i for i in target_units]
    isource_units = make_list(source_units, n=len(ntsd.columns))
    if isource_units is not None:
        isource_units = ["" if i is None else i for i in isource_units]

    # Create completely filled list of units from the column names.
    tsource_units = []
    for inx in list(range(len(ntsd.columns))):
        if isinstance(ntsd.columns[inx], (str, bytes)):
            words = ntsd.columns[inx].split(":")
            if len(words) >= 2:
                tsource_units.append(words[1])
            else:
                tsource_units.append("")
        else:
            tsource_units.append(ntsd.columns[inx])

    # Combine isource_units and tsource_units into su.
    su = []
    if isource_units is not None:
        for isource, tsource in zip(isource_units, tsource_units):
            if not tsource:
                tsource = isource
            if isource != tsource:
                raise ValueError(
                    error_wrapper(
                        f"""
The units specified by the "source_units" keyword and in the second ":"
delimited field in the DataFrame column name must match.

"source_unit" keyword is {isource_units}
Column name source units are {tsource_units}
                                                       """
                    )
                )
            su.append(tsource)
    else:
        su = [""] * len(ntsd.columns)

    if source_units_required is True and "" in su:
        raise ValueError(
            error_wrapper(
                f"""
Source units must be specified either using "source_units" keyword of in the
second ":" delimited field in the column name.  Instead you have {su}.
                                           """
            )
        )
    names = []
    for inx, unit in enumerate(su):
        if isinstance(ntsd.columns[inx], (str, bytes)):
            words = ntsd.columns[inx].split(":")
            if unit:
                tmpname = ":".join([words[0], unit])
                if len(words) > 2:
                    tmpname = tmpname + ":" + ":".join(words[2:])
                names.append(tmpname)
            else:
                names.append(":".join(words))
        else:
            names.append(ntsd.columns[inx])
    ntsd.columns = names

    if su is None and target_units is not None:
        raise ValueError(
            error_wrapper(
                f"""
To specify target_units, you must also specify source_units.  You can
specify source_units either by using the `source_units` keyword or placing
in the name of the data column as the second ':' separated field.

The `source_units` keyword must specify units that are convertible
to the `target_units`: {target_units}
"""
            )
        )

    # Convert source_units to target_units.
    if target_units is not None:
        ncolumns = []
        for inx, colname in enumerate(ntsd.columns):
            words = str(colname).split(":")
            if len(words) > 1:
                # convert words[1] to target_units[inx]
                try:
                    # Would be nice in the future to carry around units,
                    # however at the moment most tstoolbox functions will not
                    # work right with units specified.
                    # This single command uses pint to convert units and
                    # the "np.array(..., dtype=float)" command removes pint
                    # units from the converted pandas Series.
                    ntsd[str(colname)] = np.array(
                        pd.Series(
                            ntsd[str(colname)].astype(float), dtype=f"pint[{words[1]}]"
                        ).pint.to(target_units[inx]),
                        dtype=float,
                    )
                    words[1] = target_units[inx]
                except AttributeError:
                    raise ValueError(
                        error_wrapper(
                            """
No conversion between {} and {}.""".format(
                                words[1], target_units[inx]
                            )
                        )
                    )
            ncolumns.append(":".join(words))
        ntsd.columns = ncolumns

    return memory_optimize(ntsd)


def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def transform_args(**trans_func_for_arg):
    """
    Make a decorator that transforms function arguments before calling the
    function.  Works with plain functions and bounded methods.
    """

    def transform_args_decorator(func):
        sig = inspect.signature(func)

        @wraps(func)
        def transform_args_wrapper(*args, **kwargs):
            # get a {argname: argval, ...} dict from *args and **kwargs
            # Note: Didn't really need an if/else here but I am assuming
            # that getcallargs gives us an overhead that can be avoided if
            # there's only keyword args.
            val_of_argname = sig.bind(*args, **kwargs)
            val_of_argname.apply_defaults()
            for argname, trans_func in trans_func_for_arg.items():
                val_of_argname.arguments[argname] = trans_func(
                    val_of_argname.arguments[argname]
                )
            # apply transform functions to argument values
            return func(*val_of_argname.args, **val_of_argname.kwargs)

        return transform_args_wrapper

    return transform_args_decorator


@transform_args(
    pick=make_list, names=make_list, source_units=make_list, target_units=make_list
)
@typic.al
def common_kwds(
    input_tsd=None,
    start_date=None,
    end_date=None,
    pick: Optional[List[Union[int, str]]] = None,
    force_freq: Optional[str] = None,
    groupby: Optional[str] = None,
    dropna: Optional[Literal["no", "any", "all"]] = "no",
    round_index: Optional[str] = None,
    clean: bool = False,
    target_units=None,
    source_units=None,
    source_units_required: bool = False,
    bestfreq: bool = True,
    parse_dates: bool = True,
    extended_columns: bool = False,
    skiprows: Optional[Union[List[int], int]] = None,
    index_type: Literal["datetime", "number"] = "datetime",
    names: Optional[Union[List[str], str]] = None,
    usecols: Optional[List[Union[int, str]]] = None,
    por: bool = False,
):
    """Process all common_kwds across sub-commands into single function.

    Parameters
    ----------
    input_tsd: DataFrame
        Input data which should be a DataFrame.

    Returns
    -------
    df: DataFrame
        DataFrame altered according to options.

    """
    # The "por" keyword is stupid, since it is a synonym for "dropna" equal
    # to "no".  Discovered this after implementation and should deprecate
    # and remove in the future.
    if por is True:
        dropna = "no"

    ntsd = read_iso_ts(
        input_tsd,
        parse_dates=parse_dates,
        extended_columns=extended_columns,
        dropna=dropna,
        force_freq=force_freq,
        skiprows=skiprows,
        index_type=index_type,
        usecols=usecols,
        clean=clean,
    )

    if names is not None:
        ntsd.columns = names

    ntsd = _pick(ntsd, pick)

    ntsd = _normalize_units(
        ntsd, source_units, target_units, source_units_required=source_units_required
    )

    if clean is True:
        ntsd = ntsd.sort_index()
        ntsd = ntsd[~ntsd.index.duplicated()]

    ntsd = _round_index(ntsd, round_index=round_index)

    if bestfreq is True:
        ntsd = asbestfreq(ntsd, force_freq=force_freq)

    ntsd = _date_slice(ntsd, start_date=start_date, end_date=end_date, por=por)

    if ntsd.index.inferred_type == "datetime64":
        ntsd.index.name = "Datetime"

    if dropna in ["any", "all"]:
        ntsd = ntsd.dropna(axis="index", how=dropna)
    else:
        try:
            if bestfreq is True:
                ntsd = asbestfreq(ntsd, force_freq=force_freq)
        except ValueError:
            pass

    if groupby is not None:
        if groupby == "months_across_years":
            return ntsd.groupby(lambda x: x.month)
        return ntsd.resample(groupby)

    return ntsd


def _pick(tsd: DataFrame, columns: Any) -> DataFrame:
    columns = make_list(columns)
    if not columns:
        return tsd
    ncolumns = []

    for i in columns:
        if i in tsd.columns:
            # if using column names
            ncolumns.append(tsd.columns.tolist().index(i))
            continue

        if i == tsd.index.name:
            # if wanting the index
            # making it -1 that will be evaluated later...
            ncolumns.append(-1)
            continue

        # if using column numbers
        try:
            target_col = int(i) - 1
        except ValueError:
            raise ValueError(
                error_wrapper(
                    """
The name {} isn't in the list of column names
{}.
""".format(
                        i, tsd.columns
                    )
                )
            )
        if target_col < -1:
            raise ValueError(
                error_wrapper(
                    """
The requested column "{}" must be greater than or equal to 0.
First data column is 1, index is column 0.
""".format(
                        i
                    )
                )
            )
        if target_col > len(tsd.columns):
            raise ValueError(
                error_wrapper(
                    """
The request column index {} must be less than or equal to the
number of columns {}.
""".format(
                        i, len(tsd.columns)
                    )
                )
            )

        # columns names or numbers or index organized into
        # numbers in ncolumns
        ncolumns.append(target_col)

    if len(ncolumns) == 1 and ncolumns[0] != -1:
        return pd.DataFrame(tsd[tsd.columns[ncolumns]])

    newtsd = pd.DataFrame()
    for index, col in enumerate(ncolumns):
        if col == -1:
            # Use the -1 marker to indicate index
            jtsd = pd.DataFrame(tsd.index, index=tsd.index)
        else:
            try:
                jtsd = pd.DataFrame(tsd.iloc[:, col], index=tsd.index)
            except IndexError:
                jtsd = pd.DataFrame(tsd.loc[:, col], index=tsd.index)

        newtsd = newtsd.join(jtsd, lsuffix="_{}".format(index), how="outer")
    return newtsd


def _date_slice(
    input_tsd: DataFrame,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    por=False,
) -> DataFrame:
    """Private function to slice time series."""
    if input_tsd.index.inferred_type == "datetime64":
        if start_date is None:
            start_date = input_tsd.index[0]

        if end_date is None:
            end_date = input_tsd.index[-1]

        if input_tsd.index.tz is None:
            start_date = pd.Timestamp(start_date)
            end_date = pd.Timestamp(end_date)
        else:
            try:
                start_date = pd.Timestamp(start_date).tz_convert(input_tsd.index.tz)
            except TypeError:
                start_date = pd.Timestamp(start_date).tz_localize(input_tsd.index.tz)
            try:
                end_date = pd.Timestamp(end_date).tz_convert(input_tsd.index.tz)
            except TypeError:
                end_date = pd.Timestamp(end_date).tz_localize(input_tsd.index.tz)

        input_tsd = input_tsd.loc[
            (input_tsd.index >= start_date) & (input_tsd.index <= end_date), :
        ]

        if por is True:
            if start_date < input_tsd.index[0]:
                input_tsd = pd.DataFrame(index=[start_date]).append(input_tsd)
            if end_date > input_tsd.index[-1]:
                input_tsd = input_tsd.append(pd.DataFrame(index=[end_date]))
            input_tsd = asbestfreq(input_tsd)
    return input_tsd


_ANNUALS = {
    0: "DEC",
    1: "JAN",
    2: "FEB",
    3: "MAR",
    4: "APR",
    5: "MAY",
    6: "JUN",
    7: "JUL",
    8: "AUG",
    9: "SEP",
    10: "OCT",
    11: "NOV",
    12: "DEC",
}

_WEEKLIES = {0: "MON", 1: "TUE", 2: "WED", 3: "THU", 4: "FRI", 5: "SAT", 6: "SUN"}


def asbestfreq(data: DataFrame, force_freq: Optional[str] = None) -> DataFrame:
    """Test to determine best frequency to represent data.

    This uses several techniques.
    0.5.  If index is not DateTimeIndex, return
    1. If force_freq is set use .asfreq.
    2. If data.index.freq is not None, just return.
    3. If data.index.inferred_freq is set use .asfreq.
    4. Use pd.infer_freq - fails if any missing
    5. Use .is_* functions to establish A, AS, A-*, AS-*, Q, QS, M, MS
    6. Use minimum interval to establish the fixed time periods up to weekly
    7. Gives up returning None for PANDAS offset string

    """
    if not isinstance(data.index, pd.DatetimeIndex):
        return data

    if force_freq is not None:
        return data.asfreq(force_freq)

    ndiff = (
        data.index.values.astype("int64")[1:] - data.index.values.astype("int64")[:-1]
    )

    if np.any(ndiff <= 0):
        raise ValueError(
            error_wrapper(
                """
Duplicate or time reversal index entry at record {1} (start count at 0):
"{0}".

Perhaps use the "--clean" keyword on the CLI or "clean=True" if using Python or edit the
input data..
""".format(
                    data.index[:-1][ndiff <= 0][0], np.where(ndiff <= 0)[0][0] + 1
                )
            )
        )

    if data.index.freq is not None:
        return data

    # Since pandas doesn't set data.index.freq and data.index.freqstr when
    # using .asfreq, this function returns that PANDAS time offset alias code
    # also.  Not ideal at all.
    #
    # This gets most of the frequencies...
    if data.index.inferred_freq is not None:
        try:
            return data.asfreq(data.index.inferred_freq)
        except ValueError:
            pass

    # pd.infer_freq would fail if given a large dataset
    if len(data.index) > 1000:
        slic = slice(None, 999)
    else:
        slic = slice(None, None)
    try:
        infer_freq = pd.infer_freq(data.index[slic])
    except ValueError:
        infer_freq = None
    if infer_freq is not None:
        return data.asfreq(infer_freq)

    # At this point pd.infer_freq failed probably because of missing values.
    # The following algorithm would not capture things like BQ, BQS
    # ...etc.
    if np.alltrue(data.index.is_year_end):
        infer_freq = "A"
    elif np.alltrue(data.index.is_year_start):
        infer_freq = "AS"
    elif np.alltrue(data.index.is_quarter_end):
        infer_freq = "Q"
    elif np.alltrue(data.index.is_quarter_start):
        infer_freq = "QS"
    elif np.alltrue(data.index.is_month_end):
        if np.all(data.index.month == data.index[0].month):
            # Actually yearly with different ends
            infer_freq = "A-{}".format(_ANNUALS[data.index[0].month])
        else:
            infer_freq = "M"
    elif np.alltrue(data.index.is_month_start):
        if np.all(data.index.month == data.index[0].month):
            # Actually yearly with different start
            infer_freq = "A-{}".format(_ANNUALS[data.index[0].month] - 1)
        else:
            infer_freq = "MS"

    if infer_freq is not None:
        return data.asfreq(infer_freq)

    # Use the minimum of the intervals to test a new interval.
    # Should work for fixed intervals.
    ndiff = sorted(set(ndiff))
    mininterval = np.min(ndiff)
    if mininterval <= 0:
        raise ValueError
    if len(ndiff) == 1:
        ngcd = ndiff[0]
    else:
        ngcd = reduce(gcd, ndiff)
    if ngcd < 1000:
        infer_freq = "{}N".format(ngcd)
    elif ngcd < 1000000:
        infer_freq = "{}U".format(ngcd // 1000)
    elif ngcd < 1000000000:
        infer_freq = "{}L".format(ngcd // 1000000)
    elif ngcd < 60000000000:
        infer_freq = "{}S".format(ngcd // 1000000000)
    elif ngcd < 3600000000000:
        infer_freq = "{}T".format(ngcd // 60000000000)
    elif ngcd < 86400000000000:
        infer_freq = "{}H".format(ngcd // 3600000000000)
    elif ngcd < 604800000000000:
        infer_freq = "{}D".format(ngcd // 86400000000000)
    elif ngcd < 2419200000000000:
        infer_freq = "{}W".format(ngcd // 604800000000000)
        if np.all(data.index.dayofweek == data.index[0].dayofweek):
            infer_freq = infer_freq + "-{}".format(_WEEKLIES[data.index[0].dayofweek])
        else:
            infer_freq = "D"

    if infer_freq is not None:
        return data.asfreq(infer_freq)

    # Give up
    return data


def dedupIndex(
    idx: List[str], fmt: Optional[Any] = None, ignoreFirst: bool = True
) -> Index:
    # fmt:          A string format that receives two arguments:
    #               name and a counter. By default: fmt='%s.%03d'
    # ignoreFirst:  Disable/enable postfixing of first element.
    idx = pd.Series(idx)
    duplicates = idx[idx.duplicated()].unique()
    fmt = "%s.%03d" if fmt is None else fmt
    for name in duplicates:
        dups = idx == name
        ret = [
            fmt % (name, i) if (i != 0 or not ignoreFirst) else name
            for i in range(dups.sum())
        ]
        idx.loc[dups] = ret
    return pd.Index(idx)


@typic.al
def renamer(xloc: str, suffix: Optional[str] = "") -> str:
    """Print the suffix into the third ":" separated field of the header."""
    if suffix is None:
        suffix = ""
    words = xloc.split(":")
    if len(words) == 1:
        words.append("")
        words.append(suffix)
    elif len(words) == 2:
        words.append(suffix)
    elif len(words) == 3 and suffix:
        if words[2]:
            words[2] = words[2] + "_" + suffix
        else:
            words[2] = suffix
    return ":".join(words)


# Utility
def print_input(
    iftrue,
    intds,
    output,
    suffix="",
    date_format=None,
    float_format="g",
    tablefmt="csv",
    showindex="never",
):
    """Print the input time series also."""
    return printiso(
        return_input(iftrue, intds, output, suffix=suffix),
        date_format=date_format,
        float_format=float_format,
        tablefmt=tablefmt,
        showindex=showindex,
    )


def return_input(
    iftrue: Union[bool, str],
    intds: DataFrame,
    output: DataFrame,
    suffix: Optional[str] = "",
    reverse_index: bool = False,
    output_names: List = None,
) -> DataFrame:
    """Print the input time series also."""
    if output_names is None:
        output_names = []
    output.columns = output_names or [renamer(i, suffix) for i in output.columns]
    if iftrue:
        return memory_optimize(
            intds.join(output, lsuffix="_1", rsuffix="_2", how="outer")
        )
    if reverse_index is True:
        return memory_optimize(output.iloc[::-1])
    return memory_optimize(output)


def _apply_across_columns(func, xtsd, **kwds):
    """Apply a function to each column in turn."""
    for col in xtsd.columns:
        xtsd[col] = func(xtsd[col], **kwds)
    return xtsd


def _printiso(
    tsd: DataFrame,
    date_format: Optional[Any] = None,
    sep: str = ",",
    float_format: str = "g",
    showindex: str = "never",
    headers: str = "keys",
    tablefmt: str = "csv",
) -> None:
    """Separate so can use in tests."""
    if isinstance(tsd, (pd.DataFrame, pd.Series)):
        if isinstance(tsd, pd.Series):
            tsd = pd.DataFrame(tsd)

        if tsd.columns.empty:
            tsd = pd.DataFrame(index=tsd.index)

        # Not perfectly true, but likely will use showindex for indices
        # that are not time stamps.
        if showindex is True:
            if not tsd.index.name:
                tsd.index.name = "UniqueID"
        else:
            if not tsd.index.name:
                tsd.index.name = "Datetime"

        print_index = True
        if tsd.index.inferred_type in ["datetime64", "period"]:
            if not tsd.index.name:
                tsd.index.name = "Datetime"
            # Someone made the decision about the name
            # This is how I include time zone info by tacking on to the
            # index.name.
            elif "datetime" not in tsd.index.name.lower():
                tsd.index.name = "Datetime"
        else:
            # This might be overkill, but tstoolbox is for time-series.
            # Revisit if necessary.
            print_index = False

        if tsd.index.name == "UniqueID":
            print_index = False

        if showindex in ["always", "default"]:
            print_index = True

    elif isinstance(tsd, (int, float, tuple, np.ndarray)):
        tablefmt = None

    ntablefmt = None
    if tablefmt in ["csv", "tsv", "csv_nos", "tsv_nos"]:
        sep = {"csv": ",", "tsv": "\t", "csv_nos": ",", "tsv_nos": "\t"}[tablefmt]
        if isinstance(tsd, pd.DataFrame):
            try:
                tsd.to_csv(
                    sys.stdout,
                    float_format="%{}".format(float_format),
                    date_format=date_format,
                    sep=sep,
                    index=print_index,
                )
                return
            except IOError:
                return
        else:
            ntablefmt = simple_separated_format(sep)

    if tablefmt is None:
        print(str(list(tsd))[1:-1])

    if ntablefmt is None:
        all_table = tb(
            tsd,
            tablefmt=tablefmt,
            showindex=showindex,
            headers=headers,
            floatfmt=float_format,
        )
    else:
        all_table = tb(
            tsd,
            tablefmt=ntablefmt,
            showindex=showindex,
            headers=headers,
            floatfmt=float_format,
        )

    if tablefmt in ["csv_nos", "tsv_nos"]:
        print(all_table.replace(" ", ""))
    else:
        print(all_table)


# Make _printiso public.  Keep both around until convert all toolboxes.
printiso = _printiso


def open_local(filein: str) -> TextIOWrapper:
    """
    Open the given input file.

    It can decode various formats too, such as gzip and bz2.

    """
    base, ext = os.path.splitext(os.path.basename(filein))
    if ext in [".gz", ".GZ"]:
        return gzip.open(filein, "rb"), base
    if ext in [".bz", ".bz2"]:
        return bz2.BZ2File(filein, "rb"), base
    return open(filein, "r"), os.path.basename(filein)


def reduce_mem_usage(props):
    """Kept here, but was too aggressive in terms of setting the dtype.

    Not used any longer.
    """
    for col in props.columns:
        try:
            if props[col].dtype == object:  # Exclude strings
                continue
        except AttributeError:
            continue

        # make variables for Int, max and min
        mx = props[col].max()
        mn = props[col].min()

        # test if column can be converted to an integer
        try:
            asint = props[col].astype(np.int64)
            result = all((props[col] - asint) == 0)
        except ValueError:
            # Want missing values to remain missing so
            # they need to remain float.
            result = False

        # Make Integer/unsigned Integer datatypes
        if result is True:
            if mn >= 0:
                if mx < np.iinfo(np.uint8).max:
                    props[col] = props[col].astype(np.uint8)
                elif mx < np.iinfo(np.uint16).max:
                    props[col] = props[col].astype(np.uint16)
                elif mx < np.iinfo(np.uint32).max:
                    props[col] = props[col].astype(np.uint32)
                else:
                    props[col] = props[col].astype(np.uint64)
            else:
                if mn > np.iinfo(np.int8).min and mx < np.iinfo(np.int8).max:
                    props[col] = props[col].astype(np.int8)
                elif mn > np.iinfo(np.int16).min and mx < np.iinfo(np.int16).max:
                    props[col] = props[col].astype(np.int16)
                elif mn > np.iinfo(np.int32).min and mx < np.iinfo(np.int32).max:
                    props[col] = props[col].astype(np.int32)
                elif mn > np.iinfo(np.int64).min and mx < np.iinfo(np.int64).max:
                    props[col] = props[col].astype(np.int64)

        else:
            # Put here whatever I come up with for float types.
            # Algorithm problem because just looking at limits doesn't help
            # with precision.
            pass

    return props


def memory_optimize(tsd: DataFrame) -> DataFrame:
    """Convert all columns to known types.

    "convert_dtypes" replaced some code here such that the "memory_optimize"
    function might go away.  Kept in case want to add additional
    optimizations.
    """
    tsd.index = pd.Index(tsd.index, dtype=None)
    tsd = tsd.convert_dtypes()
    try:
        tsd.index.freq = pd.infer_freq(tsd.index)
    except (TypeError, ValueError):
        # TypeError: Not datetime like index
        # ValueError: Less than three rows
        pass
    return tsd


def is_valid_url(url: Union[bytes, str], qualifying: Optional[Any] = None) -> bool:
    """Return whether "url" is valid."""
    min_attributes = ("scheme", "netloc")
    qualifying = min_attributes if qualifying is None else qualifying
    token = urlparse(url)
    return all(getattr(token, qualifying_attr) for qualifying_attr in qualifying)


# @typic.al
def read_iso_ts(
    *inindat,
    dropna: Literal["no", "any", "all"] = None,
    force_freq: Optional[str] = None,
    extended_columns: bool = False,
    parse_dates: bool = True,
    skiprows: Optional[Union[int, List[int]]] = None,
    index_type: Literal["datetime", "number"] = "datetime",
    names: Optional[str] = None,
    header: Optional[str] = "infer",
    sep: Optional[str] = ",",
    index_col=0,
    usecols=None,
    **kwds,
    # skiprows: Optional[Union[int, List[int]]] = None,
    # usecols: Optional[List[Union[int, str]]] = None,
) -> pd.DataFrame:
    """Read the format printed by 'printiso' and maybe other formats.

    Parameters
    ----------
    inindat: str, bytes, StringIO, file pointer, file name, DataFrame,
           Series, tuple, list, dict

        The input data.

    Returns
    -------
    df: DataFrame
        Returns a DataFrame.

    """
    # inindat
    #
    # CLI                              API                             TRANSFORM
    # ("-",)                           "-"                             [["-"]]
    #                                                                  all columns from standard input
    #
    # ("1,2,Value",)                   [1, 2, "Value"]                 [["-", 1, 2, "Value"]]
    #                                                                  extract columns from standard input
    #
    # ("fn.csv,skiprows=4",)           ["fn.csv", {"skiprow":4}]       [["fn.csv", "skiprows=4"]]
    #                                                                  all columns from "fn.csv"
    #                                                                  skipping the 5th row
    #
    # ("fn.csv,1,4,Val",)              ["fn.csv", 1, 4, "Val"]         [["fn.csv", 1, 4, "Val"]]
    #                                                                  extract columns from fn.csv
    #
    # ("fn.wdm,1002,20000",)           ["fn.wdm", 1002, 20000]         [["fn.wdm", 1002, 20000]]
    #                                                                  WDM files MUST have DSNs
    #                                                                  extract DSNs from "fn.wdm"
    #
    # ("fn.csv,Val1,Q,4 fn.wdm,201",)  [["fn.csv", "Val1", "Q", 4], ["fn.wdm", 101, 201]]
    #                                                                  extract columns from fn.csv
    #                                                                  extract DSNs from fn.wdm
    #                                  dataframe
    #                                  series                          [[series]]A
    newkwds = {}
    # Olde global way that applies to all sources.
    # usecols = kwds.get("usecols", None)
    # skiprows = kwds.get("skiprows", None)
    # index_col = kwds.get("index_col", 0)
    # parse_dates = kwds.get("parse_dates", True)
    # index_type = kwds.get("index_type", "datetime")
    # sep = kwds.get("sep", ",")
    # header = kwds.get("header", 0)
    clean = kwds.get("clean", False)
    names = kwds.get("names")

    if not inindat:
        inindat = "-"

    # Would want this to be more generic...
    na_values = []
    for spc in range(20)[1:]:
        spcs = " " * spc
        na_values.append(spcs)
        na_values.append(spcs + "nan")

    fstr = "{1}"
    if extended_columns is True:
        fstr = "{0}.{1}"

    if names is not None:
        header = 0
        names = make_list(names)

    if index_type == "number":
        parse_dates = False

    if (
        isinstance(inindat[0], (tuple, list, pd.DataFrame, pd.Series))
        and len(inindat) == 1
    ):
        inindat = inindat[0]
    sources = make_list(inindat, sep=" ", flat=False)
    if not isinstance(sources, (list, tuple)):
        sources = [sources]
    lresult_list = []
    zones = set()
    result = pd.DataFrame()
    stdin_df = pd.DataFrame()
    for source_index, source in enumerate(sources):
        res = pd.DataFrame()
        parameters = make_list(source, sep=",")
        if isinstance(parameters, list) and len(parameters) > 0:
            fname = parameters.pop(0)
        else:
            fname = parameters
            parameters = []
        # Python API
        if isinstance(fname, pd.DataFrame):
            if len(parameters) > 0:
                res = fname[parameters]
            else:
                res = fname

        elif isinstance(fname, (pd.Series, dict)):
            res = pd.DataFrame(inindat)

        elif isinstance(fname, (tuple, list, float)):
            res = pd.DataFrame({f"values{source_index}": fname}, index=[0])

        if len(res) == 0:
            # Store keywords for each source.
            newkwds = {}

            pars = [str(i) for i in parameters]
            pars = [i for i in pars if "=" in i]
            pars = [i.split("=") for i in pars]
            for key, val in pars:
                newkwds[key] = val

            parameters = [i for i in parameters if "=" not in str(i)]

            # Command line API
            # Uses wdmtoolbox, hspfbintoolbox, or pd.read_* functions.
            fpi = None
            if fname in ["-", b"-"]:
                # if from stdin format must be the tstoolbox standard
                # pandas read_csv supports file like objects
                if "header" not in kwds:
                    kwds["header"] = 0
                header = 0
                fpi = sys.stdin
            elif isinstance(fname, (StringIO, BytesIO)):
                fpi = fname
                header = 0
            elif os.path.exists(str(fname)):
                # a local file
                # Read all wdm, hdf5, and, xls* files here
                header = "infer"
                sep = ","
                index_col = 0
                usecols = None
                fpi = fname
                _, ext = os.path.splitext(fname)
                if ext.lower() == ".wdm":
                    from wdmtoolbox import wdmtoolbox

                    nres = []
                    for par in parameters:
                        nres.append(wdmtoolbox.extract(",".join([fname] + [str(par)])))
                    res = pd.concat(nres, axis="columns")
                elif ext.lower() == ".hdf5":
                    if len(parameters) == 0:
                        res = pd.read_hdf(fpi, **newkwds)
                    else:
                        res = pd.DataFrame()
                        for i in parameters:
                            res = res.join(
                                pd.read_hdf(fname, key=i, **newkwds), how="outer"
                            )
                elif ext.lower() in [
                    ".xls",
                    ".xlsx",
                    ".xlsm",
                    ".xlsb",
                    ".odf",
                    ".ods",
                    ".odt",
                ]:
                    if len(parameters) == 0:
                        sheet = [0]
                    elif len(parameters) == 1:
                        sheet = parameters[0]
                    else:
                        sheet = make_list(parameters)

                    res = pd.read_excel(
                        fname,
                        sheet_name=sheet,
                        keep_default_na=True,
                        header=header,
                        na_values=na_values,
                        index_col=index_col,
                        usecols=usecols,
                        parse_dates=parse_dates,
                        skiprows=skiprows,
                        **newkwds,
                    )
                    if isinstance(res, dict):
                        res = pd.concat(res, axis="columns")
                        # Collapse columns MultiIndex
                        fi = res.columns.to_flat_index()
                        fi = ["_".join((str(i[0]), str(i[1]))) for i in fi]
                        res.columns = fi

            elif is_valid_url(str(fname)):
                # a url?
                header = "infer"
                fpi = fname
            elif isinstance(fname, bytes):
                # Python API
                if b"\n" in fname or b"\r" in fname:
                    # a string?
                    header = 0
                    fpi = BytesIO(fname)
                else:
                    parameters.insert(0, fname)
                    fname = "-"
                    header = 0
                    fpi = sys.stdin
            elif isinstance(fname, str):
                # Python API
                if "\n" in fname or "\r" in fname:
                    # a string?
                    header = 0
                    fpi = StringIO(fname)
                else:
                    parameters.insert(0, fname)
                    header = 0
                    fpi = sys.stdin
                    fname = "-"
            elif isinstance(fname, (StringIO, BytesIO)):
                # Python API
                header = "infer"
                fpi = fname
            else:
                # Maybe fname and parameters are actual column names of standard input.
                parameters.insert(0, fname)
                fname = "-"
                header = 0
                fpi = sys.stdin

            if len(res) == 0:
                if fname == "-" and not stdin_df.empty:
                    res = stdin_df
                else:
                    res = pd.read_csv(
                        fpi,
                        engine="python",
                        infer_datetime_format=True,
                        keep_default_na=True,
                        skipinitialspace=True,
                        header=header,
                        sep=sep,
                        na_values=na_values,
                        index_col=index_col,
                        parse_dates=parse_dates,
                        skiprows=skiprows,
                        **newkwds,
                    )
                if fname == "-" and stdin_df.empty:
                    stdin_df = res
                res = _pick(res, parameters)

        lresult_list.append(res)
        try:
            zones.add(res.index.tzinfo)
        except AttributeError:
            pass

    first = []
    second = []
    rest = []
    for col in res.columns:
        words = [i.strip() for i in str(col).split(":")]
        nwords = [i.strip("0123456789") for i in words]

        first.append(fstr.format(fname, words[0].strip()))

        if len(words) > 1:
            second.append([nwords[1]])
        else:
            second.append([])
        if len(nwords) > 2:
            rest.append(nwords[2:])
        else:
            rest.append([])
    first = [[i.strip()] for i in dedupIndex(first)]
    res.columns = [":".join(i + j + k) for i, j, k in zip(first, second, rest)]

    # tmpc = res.columns.values
    # for index, i in enumerate(res.columns):
    #    if "Unnamed:" in i:
    #        words = i.split(":")
    #        tmpc[index] = words[0].strip() + words[1].strip()
    # res.columns = tmpc
    res = memory_optimize(res)

    if res.index.inferred_type != "datetime64":
        try:
            res.set_index(0, inplace=True)
        except KeyError:
            pass
    else:
        try:
            words = res.index.name.split(":")
        except AttributeError:
            words = ""
        if len(words) > 1:
            try:
                res.index = res.index.tz_localize(words[1])
            except TypeError:
                pass
            res.index.name = "Datetime:{}".format(words[1])
        else:
            res.index.name = "Datetime"

        # res = asbestfreq(res, force_freq=force_freq)

    if dropna in ["any", "all"]:
        res.dropna(how=dropna, inplace=True)

    if len(lresult_list) > 1:
        epoch = pd.to_datetime("2000-01-01")
        moffset = epoch + to_offset("A")
        offset_set = set()
        for res in lresult_list:
            if res.index.inferred_type != "datetime64":
                continue
            if len(zones) != 1:
                try:
                    res.index = res.index.tz_convert(None)
                except (TypeError, AttributeError):
                    pass

            # Remove duplicate times if hourly and daylight savings.
            if clean is True:
                res = res.sort_index()
                res = res[~res.index.duplicated()]

            res = asbestfreq(res)
            if res.index.inferred_freq is not None and moffset > epoch + to_offset(
                res.index.inferred_freq
            ):
                moffset = epoch + to_offset(res.index.inferred_freq)
                offset_set.add(moffset)

        result = pd.DataFrame()
        for df in lresult_list:
            if len(offset_set) < 2:
                result = result.join(df, how="outer", rsuffix="_r")
            else:
                result = result.join(
                    df.asfreq(moffset - epoch), how="outer", rsuffix="_r"
                )
    else:
        result = lresult_list[0]

    # Assign names to the index and columns.
    if names is not None:
        result.index.name = names.pop(0)
        result.columns = names

    result.sort_index(inplace=True)
    return result.convert_dtypes()



# -*- coding: utf-8 -*-
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import pytest

from tstoolbox import tstoolbox

# Pull this in once.
df = tstoolbox.aggregate(
    agg_interval="D", clean=True, input_ts="tests/02234500_65_65.csv"
)
# Pull this in once.
dfa = tstoolbox.aggregate(
    agg_interval="A", clean=True, input_ts="tests/02234500_65_65.csv"
)
dfa = tstoolbox.equation(
    "x1*120@x2", input_ts=dfa, output_names=["Elevation::mean", "Flow::mean"]
)


@pytest.mark.mpl_image_compare(tolerance=6)
def test_double_mass():
    plt.close("all")
    return tstoolbox.plot(
        type="double_mass",
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_double_mass_mult():
    plt.close("all")
    return tstoolbox.plot(
        type="double_mass",
        columns=[2, 3, 3, 2],
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_double_mass_marker():
    plt.close("all")
    return tstoolbox.plot(
        type="double_mass",
        columns=[2, 3, 3, 2],
        linestyles=" ",
        markerstyles="auto",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_boxplot():
    plt.close("all")
    xdf = tstoolbox.read(
        ["tests/02234500_65_65.csv", "tests/02325000_flow.csv"],
        clean=True,
        append="combine",
    )
    return tstoolbox.plot(
        input_ts=xdf,
        clean=True,
        columns=[2, 3],
        type="boxplot",
        ofilename=None,
        plot_styles="classic",
    )


# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_scatter_matrix():
#     plt.close("all")
#     return tstoolbox.plot(
#         type="scatter_matrix",
#         clean=True,
#         input_ts="tests/02234500_65_65.csv",
#         ofilename=None,
#         plot_styles="classic",
#     )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_lag_plot():
    plt.close("all")
    return tstoolbox.plot(
        columns=1, type="lag_plot", input_ts=df, ofilename=None, plot_styles="classic"
    )


# Can't have a bootstrap test since random selections are made.
# @image_comparison(baseline_images=['bootstrap'],
#                   tol=0.019, extensions=['png'])
# def test_bootstrap():
#     return tstoolbox.plot(type='bootstrap',
#                    clean=True,
#                    columns=2,
#                    input_ts='tests/02234500_65_65.csv')


@pytest.mark.mpl_image_compare(tolerance=6)
def test_probability_density():
    plt.close("all")
    return tstoolbox.plot(
        type="probability_density",
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_bar():
    plt.close("all")
    return tstoolbox.plot(
        type="bar", input_ts=dfa, plot_styles="classic", ofilename=None
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_barh():
    plt.close("all")
    return tstoolbox.plot(
        type="barh", input_ts=dfa, plot_styles="classic", ofilename=None
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_bar_stacked():
    plt.close("all")
    return tstoolbox.plot(
        type="bar_stacked", input_ts=dfa, plot_styles="classic", ofilename=None
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_barh_stacked():
    plt.close("all")
    return tstoolbox.plot(
        type="barh_stacked", input_ts=dfa, plot_styles="classic", ofilename=None
    )



# -*- coding: utf-8 -*-
import numpy as np

from . import utils


def nse(simulated, observed):
    """
    Calculate the Nash-Sutcliffe efficiency.

    Calculates the Nash-Sutcliffe efficiency between two variables
    simulated and observed. The NSE is calculated using the
    formula:

    NSE = 1 - sum_(n=1)^N (s_n - o_n)^2 / sum_(n=1)^N (o_n - mean(r))^2

    where s is the simulated values, o is the observed values, and
    N is the total number of values in s & o. Note that s & o must
    have the same number of values.

    Nash-Sutcliffe efficiency can range from -infinity to 1. An efficiency of
    1 (E = 1) corresponds to a perfect match of modeled discharge to the
    observed data. An efficiency of 0 (E = 0) indicates that the model
    predictions are as accurate as the mean of the observed data, whereas an
    efficiency less than zero (E < 0) occurs when the observed mean is a better
    predictor than the model or, in other words, when the residual variance
    (described by the numerator in the expression above), is larger than the
    data variance (described by the denominator). Essentially, the closer the
    model efficiency is to 1, the more accurate the model is.

    The efficiency coefficient is sensitive to extreme values and might yield
    sub-optimal results when the dataset contains large outliers in it.

    Nash-Sutcliffe efficiency can be used to quantitatively describe the
    accuracy of model outputs other than discharge. This method can be used to
    describe the predictive accuracy of other models as long as there is
    observed data to compare the model results to. For example, Nash-Sutcliffe
    efficiency has been reported in scientific literature for model simulations
    of discharge, and water quality constituents such as sediment, nitrogen,
    and phosphorus loading.

    Input:
    simulated : simulated values
    observed : observed values

    Output:
    NSE : Nash-Sutcliffe Efficiency
    """
    # Check that dimensions of simulated and observed fields match
    utils.check_arrays(simulated, observed)

    # Calculate the NSE
    nse = 1 - (
        np.sum((simulated - observed) ** 2)
        / np.sum((observed - np.mean(observed)) ** 2)
    )

    return nse



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings

import mando
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils
from .convert_index import convert_index

warnings.filterwarnings("ignore")


@mando.command(formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def convert_index_to_julian(
    input_ts="-",
    epoch="julian",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    clean=False,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    skiprows=None,
):
    """DEPRECATED: Use convert_index instead.

    Will be removed in a future version of `tstoolbox`.

    Use `convert_index` in place of `convert_index_to_julian`.

    For command line::

        tstoolbox convert_index julian ...

    For Python::

        from tstoolbox import tstoolbox
        ndf = ntstoolbox.convert_index('julian', ...)
    """
    warnings.warn(
        """
*
*   DEPRECATED in favor of using `convert_index` with the 'julian'
*   option.
*
*   Will be removed in a future version of `tstoolbox`.
*
"""
    )
    return convert_index(
        "julian",
        epoch="julian",
        columns=columns,
        input_ts=input_ts,
        start_date=start_date,
        end_date=end_date,
        round_index=round_index,
        dropna=dropna,
        clean=clean,
        skiprows=skiprows,
        names=names,
        source_units=source_units,
        target_units=target_units,
        index_type=index_type,
    )



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings
from typing import List

import mando
import numpy as np
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter
from statsmodels.nonparametric.smoothers_lowess import lowess

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")


@mando.command("fit", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def fit_cli(
    method,
    lowess_frac=0.01,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    clean=False,
    round_index=None,
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Fit model to data.

    Parameters
    ----------
    method : str
        Any of 'lowess', 'linear' or list of same. The LOWESS technique is for
        vector data, like time-series, whereas the LOESS is a generalized
        technique that can be applied to multi-dimensional data.  For working
        with time-series LOESS and LOWESS are identical.
    lowess_frac : float
        [optional, default=0.01, range between 0 and 1]

        Fraction of data used for 'method'="lowess".
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${clean}
    ${round_index}
    ${skiprows}
    ${index_type}
    ${names}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        fit(
            method,
            lowess_frac=lowess_frac,
            input_ts=input_ts,
            skiprows=skiprows,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            clean=clean,
            round_index=round_index,
            index_type=index_type,
            names=names,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@tsutils.transform_args(method=tsutils.make_list)
@typic.al
@tsutils.copy_doc(fit_cli)
def fit(
    method: List[Literal["lowess", "linear"]],
    lowess_frac: tsutils.FloatBetweenZeroAndOne = 0.01,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    clean=False,
    round_index=None,
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    print_input=False,
):
    """Calculate fit."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    ntsd = pd.DataFrame()
    tmptsd = pd.DataFrame()
    for meth in method:
        if meth == "lowess":
            for cname, cdata in tsd.iteritems():
                smooth = lowess(cdata, tsd.index, frac=lowess_frac)
                index, data = np.transpose(smooth)
                ftsd = pd.DataFrame(data, index=tsd.index, columns=[cname])
                tmptsd = tmptsd.join(ftsd, how="outer")
        elif meth == "linear":
            for cname, cdata in tsd.iteritems():
                index = tsd.index.astype("l")
                index = index - index[0]
                m, b = np.polyfit(index, cdata, 1)
                data = m * index + b
                ftsd = pd.DataFrame(data, index=tsd.index, columns=[cname])
                tmptsd = tmptsd.join(ftsd, how="outer")

        tmptsd.columns = [tsutils.renamer(i, meth) for i in tmptsd.columns]
        ntsd = ntsd.join(tmptsd, how="outer")
    return tsutils.return_input(print_input, tsd, ntsd)



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings

import mando
import numpy as np
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


warnings.filterwarnings("ignore")


def lpdes(fc, t, ns):
    """Subroutine LPDES evaluates the coefficients for a low pass filter."""
    a = np.zeros(ns)
    b = np.zeros(ns)
    c = np.zeros(ns)
    wcp = np.sin(fc * np.pi * t) / np.cos(fc * np.pi * t)
    for k in range(ns):
        cs = np.cos(float(2 * (k + ns) - 1) * np.pi / float(4 * ns))
        x = 1.0 / (1.0 + wcp * wcp - 2.0 * wcp * cs)
        a[k] = wcp * wcp * x
        b[k] = 2.0 * (wcp * wcp - 1.0) * x
        c[k] = (1.0 + wcp * wcp + 2.0 * wcp * cs) * x
    return a, b, c


def hpdes(fc, t, ns):
    """Subroutine HPDES evaluates the coefficients for a high pass filter."""
    a = np.zeros(ns)
    b = np.zeros(ns)
    c = np.zeros(ns)
    pi = 3.1415926536
    wcp = np.sin(fc * pi * t) / np.cos(fc * pi * t)
    for k in range(ns):
        cs = np.cos(float(2 * (k + ns) - 1) * pi / float(4 * ns))
        a[k] = 1.0 / (1.0 + wcp * wcp - 2.0 * wcp * cs)
        b[k] = 2.0 * (wcp * wcp - 1.0) * a[k]
        c[k] = (1.0 + wcp * wcp + 2.0 * wcp * cs) * a[k]
    return a, b, c


def bpdes(f1, f2, t, ns):
    """Subroutine BPDES evaluates the coefficients for a band pass filter."""
    a = np.zeros(ns)
    b = np.zeros(ns)
    c = np.zeros(ns)
    d = np.zeros(ns)
    e = np.zeros(ns)
    pi = 3.1415926536
    w1 = np.sin(f1 * pi * t) / np.cos(f1 * pi * t)
    w2 = np.sin(f2 * pi * t) / np.cos(f2 * pi * t)
    wc = w2 - w1
    q = wc * wc + 2.0 * w1 * w2
    s = w1 * w1 * w2 * w2
    for k in range(ns):
        cs = np.cos(float(2 * (k + ns) - 1) * pi / float(4 * ns))
        p = -2.0 * wc * cs
        r = p * w1 * w2
        x = 1.0 + p + q + r + s
        a[k] = wc * wc / x
        b[k] = (-4.0 - 2.0 * p + 2.0 * r + 4.0 * s) / x
        c[k] = (6.0 - 2.0 * q + 6.0 * s) / x
        d[k] = (-4.0 + 2.0 * p - 2.0 * r + 4.0 * s) / x
        e[k] = (1.0 - p + q - r + s) / x
        return a, b, c, d, e


def _transform(vector, filter_pass, lowpass_cutoff, highpass_cutoff, window_len):
    """Private function used by FFT filtering.

    Parameters
    ----------
    vector: array_like, evenly spaced samples in time

    Returns
    -------
    vector of filtered values

    """
    import numpy.fft as F

    result = F.rfft(vector, len(vector))

    freq = F.fftfreq(len(vector))[: len(vector) // 2 + 1]
    factor = np.ones_like(freq)

    if filter_pass in ["lowpass", "bandpass"]:
        factor[freq > 1.0 / float(lowpass_cutoff)] = 0.0
    if filter_pass in ["highpass", "bandpass"]:
        factor[freq < 1.0 / float(highpass_cutoff)] = 0.0
    if filter_pass == "bandstop":
        factor[
            freq < 1.0 / float(lowpass_cutoff) and freq > 1.0 / float(highpass_cutoff)
        ] = 0.0

    factor = np.pad(
        factor,
        window_len + 1,
        mode="edge",
    )

    factor = np.convolve(factor, [1.0 / window_len] * window_len, mode=1)
    factor = factor[window_len + 1 : -(window_len + 1)]

    result = result * factor

    rvector = F.irfft(result, len(vector))

    return np.atleast_1d(rvector)


@mando.command("filter", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def filter_cli(
    filter_type,
    filter_pass,
    lowpass_cutoff=None,
    highpass_cutoff=None,
    window_len=3,
    butterworth_stages=1,
    reverse_second_stage=True,
    input_ts="-",
    start_date=None,
    end_date=None,
    columns=None,
    float_format="g",
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    round_index=None,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Apply different filters to the time-series.

    Parameters
    ----------
    filter_type: str
        OneOf("fft", "butterworth")
        The "fft" and "butterworth" types are defined by cutoff frequencies.
        The "fft" is the Fast Fourier Transform filter in the frequency domain.
    filter_pass: str
        OneOf("lowpass", "highpass", "bandpass", "bandstop")
        Indicates what frequencies to block.
    lowpass_cutoff: float
        [optional, default is None, used only if `filter_pass` equals
         "lowpass", "bandpass" or "bandstop"]

        The low frequency cutoff when `filter_pass` equals "vertical",
        "bandpass", or "bandstop".
    highpass_cutoff: float
        [optional, default is None, used only if `filter_pass` equals
         "highpass", "bandpass" or "bandstop"]

        The high frequency cutoff when `filter_pass` equals "highpass",
        "bandpass", or "bandstop".
    window_len: int
        [optional, default is 3]

        Will soften the edges of the "fft" filter in the frequency domain.
        The larger the number the softer the filter edges.  A value of 1
        will have a brick wall step function which may introduce
        frequencies into the filtered output.
    butterworth_stages: int
        [optional, default is 1]

        The order of the butterworth filter.
    reverse_second_stage: bool
        [optional, default is True]

        Will perform a second filter in reverse to eliminate shifting
        in time caused by the first filter.
    ${input_ts}
    ${start_date}
    ${end_date}
    ${columns}
    ${float_format}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${round_index}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        filter(
            filter_type,
            filter_pass,
            butterworth_stages=butterworth_stages,
            reverse_second_stage=reverse_second_stage,
            lowpass_cutoff=lowpass_cutoff,
            highpass_cutoff=highpass_cutoff,
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            print_input=print_input,
            window_len=window_len,
            source_units=source_units,
            target_units=target_units,
            round_index=round_index,
        ),
        float_format=float_format,
        tablefmt=tablefmt,
    )


@typic.constrained(ge=1)
class IntGreaterThanOrEqualToOne(int):
    """Integer greater than zero."""


@typic.constrained(ge=1, le=3)
class IntBetweenOneAndThree(int):
    """Integer between 1 and 3 inclusive."""


@typic.constrained(ge=0)
class FloatGreaterThanOrEqualToZero(float):
    """Positive float value."""


@typic.al
@tsutils.copy_doc(filter_cli)
def filter(
    filter_type: Literal[
        "flat",
        "hanning",
        "hamming",
        "bartlett",
        "blackman",
        "fft",
        "butterworth",
    ],
    filter_pass: Literal["lowpass", "highpass", "bandpass", "bandstop"],
    butterworth_stages: IntBetweenOneAndThree = 1,
    butterworth_reverse_second_stage: bool = True,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    print_input=False,
    lowpass_cutoff: FloatGreaterThanOrEqualToZero = None,
    highpass_cutoff: FloatGreaterThanOrEqualToZero = None,
    window_len: IntGreaterThanOrEqualToOne = 3,
    source_units=None,
    target_units=None,
    round_index=None,
):
    """Apply different filters to the time-series."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    if len(tsd.values) < window_len:
        raise ValueError(
            tsutils.error_wrapper(
                """
Input vector (length={}) needs to be bigger than window size ({}).
""".format(
                    len(tsd.values), window_len
                )
            )
        )

    if filter_type in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
        warnings.warn(
            tsutils.error_wrapper(
                """
DEPRECATED: The `filter_type`s "flat", "hanning", "hamming", "bartlett", and
"blackman" are implemented with greater capabilities in the "rolling_window"
function in tstoolbox.  Eventually they will be removed from the "filter" function."""
            )
        )
        # Inelegant - but works.  If any rolling_window filters then just set
        # lowpass_cutoff and highpass_cutoff to anything since it isn't used.
        if filter_pass == "lowpass":
            lowpass_cutoff = 1
            highpass_cutoff = None
        elif filter_pass == "highpass":
            lowpass_cutoff = None
            highpass_cutoff = 1
        else:
            lowpass_cutoff = 1
            highpass_cutoff = 1

    if filter_pass in ["bandpass", "bandstop"]:
        # Need both low_cutoff and highpass_cutoff for "bandpass" and "bandstop".
        if lowpass_cutoff is None or highpass_cutoff is None:
            raise ValueError(
                tsutils.error_wrapper(
                    f"""
The "bandpass" and "bandstop" options for `filter_pass` require values
for the `lowpass_cutoff` and `highpass_cutoff` keywords.  You have
"{lowpass_cutoff}" for `lowpass_cutoff` and "{highpass_cutoff}" for
`highpass_cutoff`."""
                )
            )

    if filter_pass == "lowpass":
        if lowpass_cutoff is None:
            raise ValueError(
                tsutils.error_wrapper(
                    f"""
The "lowpass" option for `filter_pass` requires a value for
`lowpass_cutoff`.  You have "{lowpass_cutoff}"."""
                )
            )
        if highpass_cutoff is not None:
            warnings.warn(
                tsutils.error_wrapper(
                    f"""
The `highpass_cutoff` value of {highpass_cutoff} is ignored it
`filter_pass` is "lowpass"."""
                )
            )

    if filter_pass == "highpass":
        if highpass_cutoff is None:
            raise ValueError(
                tsutils.error_wrapper(
                    f"""
The "highpass" option for `filter_pass` requires a value for
`highpass_cutoff`.  You have "{highpass_cutoff}"."""
                )
            )
        if lowpass_cutoff is not None:
            warnings.warn(
                tsutils.error_wrapper(
                    f"""
The `lowpass_cutoff` value of {lowpass_cutoff} is ignored it
`filter_pass` is "highpass"."""
                )
            )

    if print_input is True:
        ntsd = tsutils.asbestfreq(tsd.copy())
    else:
        ntsd = tsutils.asbestfreq(tsd)

    tdelt = pd.to_timedelta(ntsd.index.freq).delta / 86400000000000

    for col in tsd.columns:
        if filter_type == "fft":
            ntsd[col] = _transform(
                tsd[col],
                filter_pass,
                lowpass_cutoff,
                highpass_cutoff,
                window_len,
            )
        elif filter_type == "butterworth":
            rval = np.zeros(9)
            if filter_pass == "lowpass":
                if lowpass_cutoff >= 0.5 / tdelt:
                    raise ValueError(
                        tsutils.error_wrapper(
                            """
The "lowpass_cutoff" must be greater than 0.5/interval_in_days."""
                        )
                    )
                a, b, c = lpdes(lowpass_cutoff, tdelt, butterworth_stages)
            elif filter_pass == "highpass":
                if highpass_cutoff >= 0.5 / tdelt:
                    raise ValueError(
                        tsutils.error_wrapper(
                            """
The "highpass_cutoff" must be greater than 0.5/interval_in_days."""
                        )
                    )
                a, b, c = hpdes(highpass_cutoff, tdelt, butterworth_stages)
            elif filter_pass == "bandpass":
                if lowpass_cutoff >= 0.5 / tdelt or highpass_cutoff >= 0.5 / tdelt:
                    raise ValueError(
                        tsutils.error_wrapper(
                            """
The "lowpass_cutoff" and "highpass_cutoff" must be greater than 0.5/interval_in_days."""
                        )
                    )
                a, b, c, d, e = bpdes(
                    lowpass_cutoff, highpass_cutoff, tdelt, butterworth_stages
                )

            rval = np.pad(tsd[col].values, (4, 0), mode="edge")
            for k in range(butterworth_stages):
                af = a[k]
                bf = b[k]
                cf = c[k]
                if (k == 1) and (butterworth_reverse_second_stage is True):
                    af = a[0]
                    bf = b[0]
                    cf = c[0]
                if filter_pass == "bandpass":
                    df = d[k]
                    ef = e[k]
                if filter_pass == "lowpass":
                    gval = rval
                else:
                    gval = rval
                    gval[:4] = 0.0
                if filter_pass == "lowpass":
                    gval[4:] = (
                        af * (rval[4:] + 2.0 * rval[3:-1] + rval[2:-2])
                        - bf * gval[3:-1]
                        - cf * gval[2:-2]
                    )
                elif filter_pass == "highpass":
                    gval[4:] = (
                        af * (rval[4:] - 2.0 * rval[3:-1] + rval[2:-2])
                        - bf * gval[3:-1]
                        - cf * gval[2:-2]
                    )
                elif filter_pass == "bandpass":
                    gval[4:] = (
                        af * (rval[4:] - 2.0 * rval[2:-2] + rval[:-4])
                        - bf * gval[3:-1]
                        - cf * gval[2:-2]
                        - df * gval[1:-3]
                        - ef * gval[:-4]
                    )
                gval = gval[4:]
                if k + 1 != butterworth_stages:
                    if k == 0 and butterworth_reverse_second_stage is True:
                        rval = np.flip(gval)
                    else:
                        rval = gval
                elif butterworth_reverse_second_stage is True:
                    gval = np.flip(gval)
            ntsd[col] = gval

        elif filter_type in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
            if window_len < 3:
                continue
            s = np.pad(tsd[col].values, window_len // 2, "reflect")

            if filter_type == "flat":  # moving average
                w = np.ones(window_len, "d")
            else:
                w = eval("np." + filter_type + "(window_len)")
            ntsd[col] = np.convolve(w / w.sum(), s, mode="valid")

    return tsutils.return_input(print_input, tsd, ntsd, "filter")


if __name__ == "__main__":
    from tstoolbox import tstoolbox

    df = tstoolbox.read("../../tests/02325000_flow.csv")
    filt_fft_high = filter(
        "fft", "highpass", print_input=True, input_ts=df, highpass_cutoff=10
    )
    filt_fft_low = filter(
        "fft", "lowpass", print_input=True, input_ts=df, lowpass_cutoff=10
    )
    filt_butter_high = filter(
        "butterworth", "highpass", print_input=True, input_ts=df, highpass_cutoff=0.4
    )
    filt_butter_low = filter(
        "butterworth", "lowpass", print_input=True, input_ts=df, lowpass_cutoff=0.4
    )



# -*- coding: utf-8 -*-
def check_duplicate_stats(stats1, stats2, threshold=0.01):
    """
    Check two lists of paired statistics for duplicates.

    Returns a list of the pairs that agree within to <1%.

    INPUTS:
    STATS1 : List of first statistical metric, e.g. Standard Deviations
    STATS2 : List of second statistical metric, e.g. Centered Root Mean Square
             Difference

    OUTPUTS:
    DUPLICATES : List of tuples of paired statistics that are duplicates. The
                 list contains the index locations of the pairs of statistics
                 followed by their values as 2-tuples.

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Apr 23, 2017
    """
    if threshold < 1e-7:
        ValueError("threshold value must be positive: " + str(threshold))

    # Check for non-empty lists
    if len(stats1) == 0:
        ValueError("Argument stats1 is empty list!")
    elif len(stats2) == 0:
        ValueError("Argument stats2 is empty list!")

    # Check for matching list lengths
    if len(stats1) != len(stats2):
        ValueError(
            """
*
*   Arguments stats1 and stats2 have different list lengths.
*   len(stats1) = {} != len(stats2) = {}
*
*""".format(
                len(stats1), len(stats2)
            )
        )

    # Search for duplicate pairs of statistics
    duplicates = []
    n = len(stats1)
    for i in range(n):
        for j in range(i + 1, n):
            diff1 = abs((stats1[i] - stats1[j]) / stats1[i])
            diff2 = abs((stats2[i] - stats2[j]) / stats2[i])
            if diff1 < threshold and diff2 < threshold:
                duplicates.append(
                    (i, j, (stats1[i], stats2[i]), (stats1[j], stats2[j]))
                )

    return duplicates



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings
from typing import List, Union

import HydroErr as he
import mando
import numpy as np
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import skill_metrics as sm
from .. import tsutils
from .read import read

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

warnings.filterwarnings("ignore")


@mando.command("gof", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def gof_cli(
    obs_col=None,
    sim_col=None,
    stats="default",
    replace_nan=None,
    replace_inf=None,
    remove_neg=False,
    remove_zero=False,
    start_date=None,
    end_date=None,
    round_index=None,
    clean=False,
    index_type="datetime",
    source_units=None,
    target_units=None,
    tablefmt="plain",
    float_format=".3f",
    kge_sr=1.0,
    kge09_salpha=1.0,
    kge12_sgamma=1.0,
    kge_sbeta=1.0,
):
    """Will calculate goodness of fit statistics between two time-series.

    The first time series must be the observed, the second the simulated
    series.  You can only give two time-series.

    Parameters
    ----------
    obs_col
        If integer represents the column number of standard input. Can be
        If integer represents the column number of standard input. Can be
        a csv, wdm, hdf or xlsx file following format specified in
        'tstoolbox read ...'.
    sim_col
        If integer represents the column number of standard input. Can be
        a csv, wdm, hdf or xlsx file following format specified in
        'tstoolbox read ...'.
    stats: str
        [optional,  Python: list, Command line: comma separated string,
        default is 'default']

        Comma separated list of statistical measures.

        You can select two groups of statistical measures.

        +------------+---------------------------------------+
        | stats      | Description                           |
        +============+=======================================+
        | default    | A subset of common statistic measures |
        +------------+---------------------------------------+
        | all        | All available statistic measures      |
        +------------+---------------------------------------+

        The 'default' set of statistics are:

        +-----------------+--------------------------------------------------+
        | stats           | Description                                      |
        +=================+==================================================+
        | me              | Mean error or bias                               |
        |                 | -inf < ME < inf, close to 0 is better            |
        +-----------------+--------------------------------------------------+
        | pc_bias         | Percent Bias                                     |
        |                 | -inf < PC_BIAS < inf, close to 0 is better       |
        +-----------------+--------------------------------------------------+
        | apc_bias        | Absolute Percent Bias                            |
        |                 | 0 <= APC_BIAS < inf, close to 0 is better        |
        +-----------------+--------------------------------------------------+
        | rmsd            | Root Mean Square Deviation/Error                 |
        |                 | 0 <= RMSD < inf, smaller is better               |
        +-----------------+--------------------------------------------------+
        | crmsd           | Centered Root Mean Square Deviation/Error        |
        +-----------------+--------------------------------------------------+
        | corrcoef        | Pearson Correlation coefficient (r)              |
        |                 | -1 <= r <= 1                                     |
        |                 | 1 perfect positive correlation                   |
        |                 | 0 complete randomness                            |
        |                 | -1 perfect negative correlation                  |
        +-----------------+--------------------------------------------------+
        | coefdet         | Coefficient of determination (r^2)               |
        |                 | 0 <= r^2 <= 1                                    |
        |                 | 1 perfect correlation                            |
        |                 | 0 complete randomness                            |
        +-----------------+--------------------------------------------------+
        | murphyss        | Murphy Skill Score                               |
        +-----------------+--------------------------------------------------+
        | nse             | Nash-Sutcliffe Efficiency                        |
        |                 | -inf < NSE < 1, larger is better                 |
        +-----------------+--------------------------------------------------+
        | kge09           | Kling-Gupta Efficiency, 2009                     |
        |                 | -inf < KGE09 < 1, larger is better               |
        +-----------------+--------------------------------------------------+
        | kge12           | Kling-Gupta Efficiency, 2012                     |
        |                 | -inf < KGE12 < 1, larger is better               |
        +-----------------+--------------------------------------------------+
        | index_agreement | Index of agreement (d)                           |
        |                 | 0 <= d < 1, larger is better                     |
        +-----------------+--------------------------------------------------+
        | brierss         | Brier Skill Score                                |
        +-----------------+--------------------------------------------------+
        | mae             | Mean Absolute Error                              |
        |                 | 0 <= MAE < 1, larger is better                   |
        +-----------------+--------------------------------------------------+
        | mean            | observed mean, simulated mean                    |
        +-----------------+--------------------------------------------------+
        | stdev           | observed stdev, simulated stdev                  |
        +-----------------+--------------------------------------------------+

        Additional statistics:

        +-------------+-------------------------------------------------------+
        | stats       | Description                                           |
        +=============+=======================================================+
        | acc         | Anomaly correlation coefficient (ACC)                 |
        |             | -1 <= r <= 1                                          |
        |             | 1 positive correlation of variation in anomalies      |
        |             | 0 complete randomness of variation in anomalies       |
        |             | -1 negative correlation of variation in anomalies     |
        +-------------+-------------------------------------------------------+
        | d1          | Index of agreement (d1)                               |
        |             | 0 <= d1 < 1, larger is better                         |
        +-------------+-------------------------------------------------------+
        | d1_p        | Legate-McCabe Index of Agreement                      |
        |             | 0 <= d1_p < 1, larger is better                       |
        +-------------+-------------------------------------------------------+
        | d           | Index of agreement (d)                                |
        |             | 0 <= d < 1, larger is better                          |
        +-------------+-------------------------------------------------------+
        | dmod        | Modified index of agreement (dmod)                    |
        |             | 0 <= dmod < 1, larger is better                       |
        +-------------+-------------------------------------------------------+
        | drel        | Relative index of agreement (drel)                    |
        |             | 0 <= drel < 1, larger is better                       |
        +-------------+-------------------------------------------------------+
        | dr          | Refined index of agreement (dr)                       |
        |             | -1 <= dr < 1, larger is better                        |
        +-------------+-------------------------------------------------------+
        | ed          | Euclidean distance in vector space                    |
        |             | 0 <= ed < inf, smaller is better                      |
        +-------------+-------------------------------------------------------+
        | g_mean_diff | Geometric mean difference                             |
        +-------------+-------------------------------------------------------+
        | h1_mahe     | H1 absolute error                                     |
        +-------------+-------------------------------------------------------+
        | h1_mhe      | H1 mean error                                         |
        +-------------+-------------------------------------------------------+
        | h1_rmshe    | H1 root mean square error                             |
        +-------------+-------------------------------------------------------+
        | h2_mahe     | H2 mean absolute error                                |
        +-------------+-------------------------------------------------------+
        | h2_mhe      | H2 mean error                                         |
        +-------------+-------------------------------------------------------+
        | h2_rmshe    | H2 root mean square error                             |
        +-------------+-------------------------------------------------------+
        | h3_mahe     | H3 mean absolute error                                |
        +-------------+-------------------------------------------------------+
        | h3_mhe      | H3 mean error                                         |
        +-------------+-------------------------------------------------------+
        | h3_rmshe    | H3 root mean square error                             |
        +-------------+-------------------------------------------------------+
        | h4_mahe     | H4 mean absolute error                                |
        +-------------+-------------------------------------------------------+
        | h4_mhe      | H4 mean error                                         |
        +-------------+-------------------------------------------------------+
        | h4_rmshe    | H4 root mean square error                             |
        +-------------+-------------------------------------------------------+
        | h5_mahe     | H5 mean absolute error                                |
        +-------------+-------------------------------------------------------+
        | h5_mhe      | H5 mean error                                         |
        +-------------+-------------------------------------------------------+
        | h5_rmshe    | H5 root mean square error                             |
        +-------------+-------------------------------------------------------+
        | h6_mahe     | H6 mean absolute error                                |
        +-------------+-------------------------------------------------------+
        | h6_mhe      | H6 mean error                                         |
        +-------------+-------------------------------------------------------+
        | h6_rmshe    | H6 root mean square error                             |
        +-------------+-------------------------------------------------------+
        | h7_mahe     | H7 mean absolute error                                |
        +-------------+-------------------------------------------------------+
        | h7_mhe      | H7 mean error                                         |
        +-------------+-------------------------------------------------------+
        | h7_rmshe    | H7 root mean square error                             |
        +-------------+-------------------------------------------------------+
        | h8_mahe     | H8 mean absolute error                                |
        +-------------+-------------------------------------------------------+
        | h8_mhe      | H8 mean error                                         |
        +-------------+-------------------------------------------------------+
        | h8_rmshe    | H8 root mean square error                             |
        +-------------+-------------------------------------------------------+
        | h10_mahe    | H10 mean absolute error                               |
        +-------------+-------------------------------------------------------+
        | h10_mhe     | H10 mean error                                        |
        +-------------+-------------------------------------------------------+
        | h10_rmshe   | H10 root mean square error                            |
        +-------------+-------------------------------------------------------+
        | irmse       | Inertial root mean square error (IRMSE)               |
        |             | 0 <= irmse < inf, smaller is better                   |
        +-------------+-------------------------------------------------------+
        | lm_index    | Legate-McCabe Efficiency Index                        |
        |             | 0 <= lm_index < 1, larger is better                   |
        +-------------+-------------------------------------------------------+
        | maape       | Mean Arctangent Absolute Percentage Error (MAAPE)     |
        |             | 0 <= maape < pi/2, smaller is better                  |
        +-------------+-------------------------------------------------------+
        | male        | Mean absolute log error                               |
        |             | 0 <= male < inf, smaller is better                    |
        +-------------+-------------------------------------------------------+
        | mapd        | Mean absolute percentage deviation (MAPD)             |
        +-------------+-------------------------------------------------------+
        | mape        | Mean absolute percentage error (MAPE)                 |
        |             | 0 <= mape < inf, 0 indicates perfect correlation      |
        +-------------+-------------------------------------------------------+
        | mase        | Mean absolute scaled error                            |
        +-------------+-------------------------------------------------------+
        | mb_r        | Mielke-Berry R value (MB R)                           |
        |             | 0 <= mb_r < 1, larger is better                       |
        +-------------+-------------------------------------------------------+
        | mdae        | Median absolute error (MdAE)                          |
        |             | 0 <= mdae < inf, smaller is better                    |
        +-------------+-------------------------------------------------------+
        | mde         | Median error (MdE)                                    |
        |             | -inf < mde < inf, closer to zero is better            |
        +-------------+-------------------------------------------------------+
        | mdse        | Median squared error (MdSE)                           |
        |             | 0 < mde < inf, closer to zero is better               |
        +-------------+-------------------------------------------------------+
        | mean_var    | Mean variance                                         |
        +-------------+-------------------------------------------------------+
        | me          | Mean error                                            |
        |             | -inf < me < inf, closer to zero is better             |
        +-------------+-------------------------------------------------------+
        | mle         | Mean log error                                        |
        |             | -inf < mle < inf, closer to zero is better            |
        +-------------+-------------------------------------------------------+
        | mse         | Mean squared error                                    |
        |             | 0 <= mse < inf, smaller is better                     |
        +-------------+-------------------------------------------------------+
        | msle        | Mean squared log error                                |
        |             | 0 <= msle < inf, smaller is better                    |
        +-------------+-------------------------------------------------------+
        | ned         | Normalized Euclidian distance in vector space         |
        |             | 0 <= ned < inf, smaller is better                     |
        +-------------+-------------------------------------------------------+
        | nrmse_iqr   | IQR normalized root mean square error                 |
        |             | 0 <= nrmse_iqr < inf, smaller is better               |
        +-------------+-------------------------------------------------------+
        | nrmse_mean  | Mean normalized root mean square error                |
        |             | 0 <= nrmse_mean < inf, smaller is better              |
        +-------------+-------------------------------------------------------+
        | nrmse_range | Range normalized root mean square error               |
        |             | 0 <= nrmse_range < inf, smaller is better             |
        +-------------+-------------------------------------------------------+
        | nse_mod     | Modified Nash-Sutcliffe efficiency (NSE mod)          |
        |             | -inf < nse_mod < 1, larger is better                  |
        +-------------+-------------------------------------------------------+
        | nse_rel     | Relative Nash-Sutcliffe efficiency (NSE rel)          |
        |             | -inf < nse_mod < 1, larger is better                  |
        +-------------+-------------------------------------------------------+
        | rmse        | Root mean square error                                |
        |             | 0 <= rmse < inf, smaller is better                    |
        +-------------+-------------------------------------------------------+
        | rmsle       | Root mean square log error                            |
        |             | 0 <= rmsle < inf, smaller is better                   |
        +-------------+-------------------------------------------------------+
        | sa          | Spectral Angle (SA)                                   |
        |             | -pi/2 <= sa < pi/2, closer to 0 is better             |
        +-------------+-------------------------------------------------------+
        | sc          | Spectral Correlation (SC)                             |
        |             | -pi/2 <= sc < pi/2, closer to 0 is better             |
        +-------------+-------------------------------------------------------+
        | sga         | Spectral Gradient Angle (SGA)                         |
        |             | -pi/2 <= sga < pi/2, closer to 0 is better            |
        +-------------+-------------------------------------------------------+
        | sid         | Spectral Information Divergence (SID)                 |
        |             | -pi/2 <= sid < pi/2, closer to 0 is better            |
        +-------------+-------------------------------------------------------+
        | smape1      | Symmetric Mean Absolute Percentage Error (1) (SMAPE1) |
        |             | 0 <= smape1 < 100, smaller is better                  |
        +-------------+-------------------------------------------------------+
        | smape2      | Symmetric Mean Absolute Percentage Error (2) (SMAPE2) |
        |             | 0 <= smape2 < 100, smaller is better                  |
        +-------------+-------------------------------------------------------+
        | spearman_r  | Spearman rank correlation coefficient                 |
        |             | -1 <= spearman_r <= 1                                 |
        |             | 1 perfect positive correlation                        |
        |             | 0 complete randomness                                 |
        |             | -1 perfect negative correlation                       |
        +-------------+-------------------------------------------------------+
        | ve          | Volumetric Efficiency (VE)                            |
        |             | 0 <= ve < 1, smaller is better                        |
        +-------------+-------------------------------------------------------+
        | watt_m      | Watterson's M (M)                                     |
        |             | -1 <= watt_m < 1, larger is better                    |
        +-------------+-------------------------------------------------------+
    replace_nan: float
        If given, indicates which value to replace NaN values with in the two
        arrays. If None, when a NaN value is found at the i-th position in the
        observed OR simulated array, the i-th value of the observed and
        simulated array are removed before the computation.
    replace_inf: float
        If given, indicates which value to replace Inf values with in the two
        arrays. If None, when an inf value is found at the i-th position in the
        observed OR simulated array, the i-th value of the observed and
        simulated array are removed before the computation.
    remove_neg: boolean
        If True, when a negative value is found at the i-th position in the
        observed OR simulated array, the i-th value of the observed AND
        simulated array are removed before the computation.
    remove_zero: boolean
        If true, when a zero value is found at the i-th position in the
        observed OR simulated array, the i-th value of the observed AND
        simulated array are removed before the computation.
    ${start_date}
    ${end_date}
    ${round_index}
    ${clean}
    ${index_type}
    ${source_units}
    ${target_units}
    ${tablefmt}
    ${float_format}
    kge_sr: float
        [optional, defaults to 1.0]

        Scaling factor for `kge09` and `kge12` correlation.
    kge09_salpha: float
        [optional, defaults to 1.0]

        Scaling factor for `kge09` alpha.
    kge12_sgamma: float
        [optional, defaults to 1.0]

        Scaling factor for `kge12` beta.
    kge_sbeta: float
        [optional, defaults to 1.0]

        Scaling factor for `kge09` and `kge12` beta.
    """
    obs_col = obs_col or 1
    sim_col = sim_col or 2
    tsutils.printiso(
        gof(
            obs_col=obs_col,
            sim_col=sim_col,
            stats=stats,
            replace_nan=replace_nan,
            replace_inf=replace_inf,
            remove_neg=remove_neg,
            remove_zero=remove_zero,
            start_date=start_date,
            end_date=end_date,
            round_index=round_index,
            clean=clean,
            index_type=index_type,
            source_units=source_units,
            target_units=target_units,
            kge_sr=kge_sr,
            kge09_salpha=kge09_salpha,
            kge12_sgamma=kge12_sgamma,
            kge_sbeta=kge_sbeta,
        ),
        tablefmt=tablefmt,
        headers=["Statistic", "Comparison", "Observed", "Simulated"],
        float_format=float_format,
    )


stats_dict = {
    "bias": ["Mean error or bias", he.me],
    "pc_bias": ["Percent bias", sm.pc_bias],
    "apc_bias": ["Absolute percent bias", sm.apc_bias],
    "rmsd": ["Root-mean-square Deviation/Error (RMSD)", he.rmse],
    "crmsd": ["Centered RMSD (CRMSD)", sm.centered_rms_dev],
    "corrcoef": ["Pearson coefficient of correlation (r)", he.pearson_r],
    "coefdet": ["Coefficient of determination (r^2)", he.r_squared],
    "murphyss": ["Skill score (Murphy)", sm.skill_score_murphy],
    "nse": ["Nash-Sutcliffe Efficiency", he.nse],
    "kge09": ["Kling-Gupta efficiency (2009)", he.kge_2009],
    "kge12": ["Kling-Gupta efficiency (2012)", he.kge_2012],
    "index_agreement": ["Index of agreement", he.d],
    "brierss": ["Brier's Score", lambda sim, obs: np.sum(sim - obs) ** 2 / len(sim)],
    "mae": ["Mean Absolute Error", he.mae],
    "mean": ["Mean", lambda x: x],
    "stdev": ["Standard deviation", lambda x: x],
    "acc": ["Anomaly correlation coefficient (ACC)", he.acc],
    "d1": ["Index of agreement (d1)", he.d1],
    "d1_p": ["Legate-McCabe Index of Agreement", he.d1_p],
    "d": ["Index of agreement (d)", he.d],
    "dmod": ["Modified index of agreement (dmod)", he.dmod],
    "drel": ["Relative index of agreement (drel)", he.drel],
    "dr": ["refined index of agreement (dr)", he.dr],
    "ed": ["Euclidean distance in vector space", he.ed],
    "g_mean_diff": ["Geometric mean difference", he.g_mean_diff],
    "h10_mahe": ["H10 mean absolute error", he.h10_mahe],
    "h10_mhe": ["H10 mean error", he.h10_mhe],
    "h10_rmshe": ["H10 root mean square error", he.h10_rmshe],
    "h1_mahe": ["H1 absolute error", he.h1_mahe],
    "h1_mhe": ["H1 mean error", he.h1_mhe],
    "h1_rmshe": ["H1 root mean square error", he.h1_rmshe],
    "h2_mahe": ["H2 mean absolute error", he.h2_mahe],
    "h2_mhe": ["H2 mean error", he.h2_mhe],
    "h2_rmshe": ["H2 root mean square error", he.h2_rmshe],
    "h3_mahe": ["H3 mean absolute error", he.h3_mahe],
    "h3_mhe": ["H3 mean error", he.h3_mhe],
    "h3_rmshe": ["H3 root mean square error", he.h3_rmshe],
    "h4_mahe": ["H4 mean absolute error", he.h4_mahe],
    "h4_mhe": ["H4 mean error", he.h4_mhe],
    "h4_rmshe": ["H4 mean error", he.h4_rmshe],
    "h5_mahe": ["H5 mean absolute error", he.h5_mahe],
    "h5_mhe": ["H5 mean error", he.h5_mhe],
    "h5_rmshe": ["H5 root mean square error", he.h5_rmshe],
    "h6_mahe": ["H6 mean absolute error", he.h6_mahe],
    "h6_mhe": ["H6 mean error", he.h6_mhe],
    "h6_rmshe": ["H6 root mean square error", he.h6_rmshe],
    "h7_mahe": ["H7 mean absolute error", he.h7_mahe],
    "h7_mhe": ["H7 mean error", he.h7_mhe],
    "h7_rmshe": ["H7 root mean square error", he.h7_rmshe],
    "h8_mahe": ["H8 mean absolute error", he.h8_mahe],
    "h8_mhe": ["H8 mean error", he.h8_mhe],
    "h8_rmshe": ["H8 root mean square error", he.h8_rmshe],
    "irmse": ["Inertial root mean square error (IRMSE)", he.irmse],
    "lm_index": ["Legate-McCabe Efficiency Index", he.lm_index],
    "maape": ["Mean Arctangent Absolute Percentage Error (MAAPE)", he.maape],
    "male": ["Mean absolute log error", he.male],
    "mapd": ["Mean absolute percentage deviation (MAPD)", he.mapd],
    "mape": ["Mean absolute percentage error (MAPE)", he.mape],
    "mase": ["Mean absolute scaled error", he.mase],
    "mb_r": ["Mielke-Berry R value (MB R)", he.mb_r],
    "mdae": ["Median absolute error (MdAE)", he.mdae],
    "mde": ["Median error (MdE)", he.mde],
    "mdse": ["Median squared error (MdSE)", he.mdse],
    "mean_var": ["Mean variance", he.mean_var],
    "me": ["Mean error or bias", he.me],
    "mle": ["Mean log error", he.mle],
    "mse": ["Mean squared error", he.mse],
    "msle": ["Mean squared log error", he.msle],
    "ned": ["Normalized Euclidian distance in vector space", he.ned],
    "nrmse_iqr": ["IQR normalized root mean square error", he.nrmse_iqr],
    "nrmse_mean": ["Mean normalized root mean square error", he.nrmse_mean],
    "nrmse_range": ["Range normalized root mean square error", he.nrmse_range],
    "nse_mod": ["Modified Nash-Sutcliffe efficiency (NSE mod)", he.nse_mod],
    "nse_rel": ["Relative Nash-Sutcliffe efficiency (NSE rel)", he.nse_rel],
    "rmse": ["Root mean square error", he.rmse],
    "rmsle": ["Root mean square log error", he.rmsle],
    "sa": ["Spectral Angle (SA)", he.sa],
    "sc": ["Spectral Correlation (SC)", he.sc],
    "sga": ["Spectral Gradient Angle (SGA)", he.sga],
    "sid": ["Spectral Information Divergence (SID)", he.sid],
    "smape1": ["Symmetric Mean Absolute Percentage Error (1) (SMAPE1)", he.smape1],
    "smape2": ["Symmetric Mean Absolute Percentage Error (2) (SMAPE2)", he.smape2],
    "spearman_r": ["Spearman rank correlation coefficient", he.spearman_r],
    "ve": ["Volumetric Efficiency (VE)", he.ve],
    "watt_m": ["Watterson’s M (M)", he.watt_m],
}


@tsutils.transform_args(stats=tsutils.make_list)
@typic.al
@tsutils.copy_doc(gof_cli)
def gof(
    obs_col=1,
    sim_col=2,
    stats: List[
        Literal[
            "default",
            "all",
            "bias",
            "pc_bias",
            "apc_bias",
            "rmsd",
            "crmsd",
            "corrcoef",
            "coefdet",
            "murphyss",
            "nse",
            "kge09",
            "kge12",
            "index_agreement",
            "brierss",
            "mae",
            "mean",
            "stdev",
            "acc",
            "d1",
            "d1_p",
            "d",
            "dmod",
            "drel",
            "dr",
            "ed",
            "g_mean_diff",
            "h10_mahe",
            "h10_mhe",
            "h10_rmshe",
            "h1_mahe",
            "h1_mhe",
            "h1_rmshe",
            "h2_mahe",
            "h2_mhe",
            "h2_rmshe",
            "h3_mahe",
            "h3_mhe",
            "h3_rmshe",
            "h4_mahe",
            "h4_mhe",
            "h4_rmshe",
            "h5_mahe",
            "h5_mhe",
            "h5_rmshe",
            "h6_mahe",
            "h6_mhe",
            "h6_rmshe",
            "h7_mahe",
            "h7_mhe",
            "h7_rmshe",
            "h8_mahe",
            "h8_mhe",
            "h8_rmshe",
            "irmse",
            "lm_index",
            "maape",
            "male",
            "mapd",
            "mape",
            "mase",
            "mb_r",
            "mdae",
            "mde",
            "mdse",
            "mean_var",
            "me",
            "mle",
            "mse",
            "msle",
            "ned",
            "nrmse_iqr",
            "nrmse_mean",
            "nrmse_range",
            "nse_mod",
            "nse_rel",
            "rmse",
            "rmsle",
            "sa",
            "sc",
            "sga",
            "sid",
            "smape1",
            "smape2",
            "spearman_r",
            "ve",
            "watt_m",
        ],
    ] = "default",
    replace_nan=None,
    replace_inf=None,
    remove_neg=False,
    remove_zero=False,
    start_date=None,
    end_date=None,
    round_index=None,
    clean=False,
    index_type="datetime",
    source_units=None,
    target_units=None,
    kge_sr: float = 1.0,
    kge09_salpha: float = 1.0,
    kge12_sgamma: float = 1.0,
    kge_sbeta: float = 1.0,
):
    """Will calculate goodness of fit statistics between two time-series."""
    if "default" in stats:
        stats = [
            "me",
            "pc_bias",
            "apc_bias",
            "rmsd",
            "crmsd",
            "corrcoef",
            "coefdet",
            "murphyss",
            "nse",
            "kge09",
            "kge12",
            "index_agreement",
            "brierss",
            "mae",
            "mean",
            "stdev",
        ]
    elif "all" in stats:
        stats = stats_dict.keys()

    # Use dropna='no' to get the lengths of both time-series.
    tsd = read(
        obs_col,
        sim_col,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        round_index=round_index,
        dropna="no",
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )
    if len(tsd.columns) != 2:
        raise ValueError(
            tsutils.error_wrapper(
                """
The "gof" requires only two time-series, the first one is the observed values
and the second is the simulated.  """
            )
        )
    lennao, lennas = tsd.isna().sum()

    tsd = tsd.dropna(how="any")

    statval = []

    obs = tsd.iloc[:, 0].astype("float64")
    sim = tsd.iloc[:, 1].astype("float64")

    nstats = [i for i in stats if i not in ["mean", "stdev"]]

    for stat in nstats:
        extra_args = {
            "replace_nan": replace_nan,
            "replace_inf": replace_inf,
            "remove_neg": remove_neg,
            "remove_zero": remove_zero,
        }
        if stat in ["crmsd", "murphyss", "brierss", "pc_bias", "apc_bias"]:
            extra_args = {}
        proc = stats_dict[stat]
        if "kge09" == stat:
            extra_args = {"s": (kge_sr, kge09_salpha, kge_sbeta)}
        if "kge12" == stat:
            extra_args = {"s": (kge_sr, kge12_sgamma, kge_sbeta)}
        statval.append(
            [
                proc[0],
                proc[1](sim, obs, **extra_args),
            ]
        )

    statval.append(["Common count observed and simulated", len(tsd.index)])

    statval.append(["Count of NaNs", None, lennao, lennas])

    if "mean" in stats:
        statval.append(["Mean", None, obs.mean(), sim.mean()])

    if "stdev" in stats:
        statval.append(["Standard deviation", None, obs.std(), sim.std()])

    return statval



# -*- coding: utf-8 -*-
def get_target_diagram_options(**kwargs):
    """
    Get optional arguments for target_diagram function.

    Retrieves the optional arguments supplied to the TARGET_DIAGRAM
    function as a variable-length keyword argument list (*KWARGS), and
    returns the values in an OPTION dictionary. Default values are
    assigned to selected optional arguments. The function will terminate
    with an error if an unrecognized optional argument is supplied.

    INPUTS:
    *kwargs : variable-length keyword argument list. The keywords by
              definition are dictionaries with keys that must correspond to
              one choices given in OUTPUTS below.

    OUTPUTS:
    option : dictionary containing option values. (Refer to
             display_target_diagram_options function for more information.)
    option['alpha']           : blending of symbol face color (0.0
                                transparent through 1.0 opaque). (Default
                                : 1.0)
    option['axismax']         : maximum for the Bias & uRMSD axis
    option['circlelinespec']  : circle line specification (default dashed
                             black, '--k')
    option['circlelinewidth'] : circle line width specification (default 0.5)
    option['circles']         : radii of circles to draw to indicate
                                isopleths of standard deviation (empty by
                                default)
    option['colormap']        : 'on'/'off' switch to map color shading of
                                 markers to colormap ('on') or min to max range
                                 of RMSDz values ('off'). Set to same value as
                                 option['nonrmsdz'].
    option['equalAxes']       : 'on'/'off' switch to set axes to be equal
    option['markerdisplayed'] : markers to use for individual experiments
    option['markerlabel']     : name of the experiment to use for marker
    option['markerlabelcolor'] : marker label color (Default 'k')
    option['markerlegend']    : 'on'/'off' switch to display marker legend
                                (Default 'off')
    option['markersize']      : marker size (Default 10)

    option['nonrmsdz']        : 'on'/'off' switch indicating values in RMSDz
                                do not correspond to total RMS Differences.
                                (Default 'off')
    option['normalized']      : statistics supplied are normalized with
                                respect to the standard deviation of reference
                                values (Default 'off')
    option['obsUncertainty']  : Observational Uncertainty (default of 0)
    option['overlay']         : 'on'/'off' switch to overlay current
                                statistics on Taylor diagram (Default 'off')
                                Only markers will be displayed.
    option['ticks']           : define tick positions (default is that used
                                by the axis function)
    option['titlecolorbar']   : title for the colorbar
    option['xticklabelpos']   : position of the tick labels along the x-axis
                                (empty by default)
    option['yticklabelpos']   : position of the tick labels along the y-axis
                                (empty by default)

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Nov 25, 2016
    """
    from . import check_on_off

    nargin = len(kwargs)

    # Set default parameters for all options
    option = {}
    option["alpha"] = 1.0
    option["axismax"] = 0.0
    option["circlelinespec"] = "k--"
    option["circlelinewidth"] = 1.5
    option["circles"] = []
    option["colormap"] = "off"
    option["equalaxes"] = "on"

    option["markercolor"] = "r"
    option["markerdisplayed"] = "marker"
    option["markerlabel"] = ""
    option["markerlabelcolor"] = "k"
    option["markerlegend"] = "off"
    option["markersize"] = 10

    option["nonrmsdz"] = "off"
    option["normalized"] = "off"
    option["obsuncertainty"] = 0.0
    option["overlay"] = "off"
    option["ticks"] = []
    option["titlecolorbar"] = ""
    option["xticklabelpos"] = []
    option["yticklabelpos"] = []
    if nargin == 0:
        # No options requested, so return with only defaults
        return option

    # Check for valid keys and values in dictionary
    for optname, optvalue in kwargs.items():
        optname = optname.lower()
        if optname not in option:
            raise ValueError("Unrecognized option: " + optname)
        # Replace option value with that from arguments
        option[optname] = optvalue

        # Check values for specific options
        if optname == "equalaxes":
            option["equalaxes"] = check_on_off(option["equalaxes"])
        elif optname == "markerlegend":
            option["markerlegend"] = check_on_off(option["markerlegend"])
        elif optname == "nonrmsdz":
            option["nonrmsdz"] = check_on_off(option["nonrmsdz"])
        elif optname == "normalized":
            option["normalized"] = check_on_off(option["normalized"])
        elif optname == "overlay":
            option["overlay"] = check_on_off(option["overlay"])

    option["colormap"] = option["nonrmsdz"]

    return option
